# emacs: -*- mode: python-mode; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the NiBabel package for the
#   copyright and license terms.
#
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##

from os.path import dirname
from os.path import join as pjoin

import numpy as np
import pytest
from numpy.testing import assert_array_almost_equal
from packaging.version import Version

import nibabel as nib
from nibabel import cifti2 as ci
from nibabel.cifti2.parse_cifti2 import _Cifti2AsNiftiHeader
from nibabel.tests import test_nifti2 as tn2
from nibabel.tests.nibabel_data import get_nibabel_data, needs_nibabel_data
from nibabel.tmpdirs import InTemporaryDirectory

NIBABEL_TEST_DATA = pjoin(dirname(nib.__file__), 'tests', 'data')
NIFTI2_DATA = pjoin(NIBABEL_TEST_DATA, 'example_nifti2.nii.gz')

CIFTI2_DATA = pjoin(get_nibabel_data(), 'nitest-cifti2')

DATA_FILE1 = pjoin(CIFTI2_DATA, '')
DATA_FILE2 = pjoin(CIFTI2_DATA, 'Conte69.MyelinAndCorrThickness.32k_fs_LR.dscalar.nii')
DATA_FILE3 = pjoin(CIFTI2_DATA, 'Conte69.MyelinAndCorrThickness.32k_fs_LR.dtseries.nii')
DATA_FILE4 = pjoin(CIFTI2_DATA, 'Conte69.MyelinAndCorrThickness.32k_fs_LR.ptseries.nii')
DATA_FILE5 = pjoin(CIFTI2_DATA, 'Conte69.parcellations_VGD11b.32k_fs_LR.dlabel.nii')
DATA_FILE6 = pjoin(CIFTI2_DATA, 'ones.dscalar.nii')
datafiles = [DATA_FILE2, DATA_FILE3, DATA_FILE4, DATA_FILE5, DATA_FILE6]


def test_space_separated_affine():
    ci.Cifti2Image.from_filename(pjoin(NIBABEL_TEST_DATA, 'row_major.dconn.nii'))


def test_read_nifti2():
    # Error trying to read a CIFTI-2 image from a NIfTI2-only image.
    filemap = ci.Cifti2Image.make_file_map()
    for k in filemap:
        filemap[k].fileobj = open(NIFTI2_DATA)
    with pytest.raises(ValueError):
        ci.Cifti2Image.from_file_map(filemap)


@needs_nibabel_data('nitest-cifti2')
def test_read_internal():
    img2 = ci.load(DATA_FILE6)
    assert isinstance(img2.header, ci.Cifti2Header)
    assert img2.shape == (1, 91282)


@needs_nibabel_data('nitest-cifti2')
def test_read_and_proxies():
    img2 = nib.load(DATA_FILE6)
    assert isinstance(img2.header, ci.Cifti2Header)
    assert img2.shape == (1, 91282)
    # While we cannot reshape arrayproxies, all images are in-memory
    assert not img2.in_memory
    data = img2.get_fdata()
    assert data is not img2.dataobj
    # Uncaching has no effect, images are always array images
    img2.uncache()
    assert data is not img2.get_fdata()


@needs_nibabel_data('nitest-cifti2')
def test_version():
    for dat in datafiles:
        img = nib.load(dat)
        assert Version(img.header.version) == Version('2')


@needs_nibabel_data('nitest-cifti2')
def test_readwritedata():
    with InTemporaryDirectory():
        for name in datafiles:
            img = ci.load(name)
            ci.save(img, 'test.nii')
            img2 = ci.load('test.nii')
            assert len(img.header.matrix) == len(img2.header.matrix)
            # Order should be preserved in load/save
            for mim1, mim2 in zip(img.header.matrix, img2.header.matrix):
                named_maps1 = [m_ for m_ in mim1 if isinstance(m_, ci.Cifti2NamedMap)]
                named_maps2 = [m_ for m_ in mim2 if isinstance(m_, ci.Cifti2NamedMap)]
                assert len(named_maps1) == len(named_maps2)
                for map1, map2 in zip(named_maps1, named_maps2):
                    assert map1.map_name == map2.map_name
                    if map1.label_table is None:
                        assert map2.label_table is None
                    else:
                        assert len(map1.label_table) == len(map2.label_table)

            assert_array_almost_equal(img.dataobj, img2.dataobj)


@needs_nibabel_data('nitest-cifti2')
def test_nibabel_readwritedata():
    with InTemporaryDirectory():
        for name in datafiles:
            img = nib.load(name)
            nib.save(img, 'test.nii')
            img2 = nib.load('test.nii')
            assert len(img.header.matrix) == len(img2.header.matrix)
            # Order should be preserved in load/save
            for mim1, mim2 in zip(img.header.matrix, img2.header.matrix):
                named_maps1 = [m_ for m_ in mim1 if isinstance(m_, ci.Cifti2NamedMap)]
                named_maps2 = [m_ for m_ in mim2 if isinstance(m_, ci.Cifti2NamedMap)]
                assert len(named_maps1) == len(named_maps2)
                for map1, map2 in zip(named_maps1, named_maps2):
                    assert map1.map_name == map2.map_name
                    if map1.label_table is None:
                        assert map2.label_table is None
                    else:
                        assert len(map1.label_table) == len(map2.label_table)
            assert_array_almost_equal(img.dataobj, img2.dataobj)


@needs_nibabel_data('nitest-cifti2')
def test_cifti2types():
    """Check that we instantiate Cifti2 classes correctly, and that our
    test files exercise all classes"""
    counter = {
        ci.Cifti2LabelTable: 0,
        ci.Cifti2Label: 0,
        ci.Cifti2NamedMap: 0,
        ci.Cifti2Surface: 0,
        ci.Cifti2VoxelIndicesIJK: 0,
        ci.Cifti2Vertices: 0,
        ci.Cifti2Parcel: 0,
        ci.Cifti2TransformationMatrixVoxelIndicesIJKtoXYZ: 0,
        ci.Cifti2Volume: 0,
        ci.Cifti2VertexIndices: 0,
        ci.Cifti2BrainModel: 0,
        ci.Cifti2MatrixIndicesMap: 0,
    }

    for name in datafiles:
        hdr = ci.load(name).header
        # Matrix and MetaData aren't conditional, so don't bother counting
        assert isinstance(hdr.matrix, ci.Cifti2Matrix)
        assert isinstance(hdr.matrix.metadata, ci.Cifti2MetaData)
        for mim in hdr.matrix:
            assert isinstance(mim, ci.Cifti2MatrixIndicesMap)
            counter[ci.Cifti2MatrixIndicesMap] += 1
            for map_ in mim:
                print(map_)
                if isinstance(map_, ci.Cifti2BrainModel):
                    counter[ci.Cifti2BrainModel] += 1
                    if isinstance(map_.vertex_indices, ci.Cifti2VertexIndices):
                        counter[ci.Cifti2VertexIndices] += 1
                    if isinstance(map_.voxel_indices_ijk, ci.Cifti2VoxelIndicesIJK):
                        counter[ci.Cifti2VoxelIndicesIJK] += 1
                elif isinstance(map_, ci.Cifti2NamedMap):
                    counter[ci.Cifti2NamedMap] += 1
                    assert isinstance(map_.metadata, ci.Cifti2MetaData)
                    if isinstance(map_.label_table, ci.Cifti2LabelTable):
                        counter[ci.Cifti2LabelTable] += 1
                        for label in map_.label_table:
                            assert isinstance(map_.label_table[label], ci.Cifti2Label)
                            counter[ci.Cifti2Label] += 1
                elif isinstance(map_, ci.Cifti2Parcel):
                    counter[ci.Cifti2Parcel] += 1
                    if isinstance(map_.voxel_indices_ijk, ci.Cifti2VoxelIndicesIJK):
                        counter[ci.Cifti2VoxelIndicesIJK] += 1
                    assert isinstance(map_.vertices, list)
                    for vtcs in map_.vertices:
                        assert isinstance(vtcs, ci.Cifti2Vertices)
                        counter[ci.Cifti2Vertices] += 1
                elif isinstance(map_, ci.Cifti2Surface):
                    counter[ci.Cifti2Surface] += 1
                elif isinstance(map_, ci.Cifti2Volume):
                    counter[ci.Cifti2Volume] += 1
                    if isinstance(
                        map_.transformation_matrix_voxel_indices_ijk_to_xyz,
                        ci.Cifti2TransformationMatrixVoxelIndicesIJKtoXYZ,
                    ):
                        counter[ci.Cifti2TransformationMatrixVoxelIndicesIJKtoXYZ] += 1

            assert list(mim.named_maps) == [m_ for m_ in mim if isinstance(m_, ci.Cifti2NamedMap)]
            assert list(mim.surfaces) == [m_ for m_ in mim if isinstance(m_, ci.Cifti2Surface)]
            assert list(mim.parcels) == [m_ for m_ in mim if isinstance(m_, ci.Cifti2Parcel)]
            assert list(mim.brain_models) == [
                m_ for m_ in mim if isinstance(m_, ci.Cifti2BrainModel)
            ]
            assert ([mim.volume] if mim.volume else []) == [
                m_ for m_ in mim if isinstance(m_, ci.Cifti2Volume)
            ]

    for klass, count in counter.items():
        assert count > 0, 'No exercise of ' + klass.__name__


@needs_nibabel_data('nitest-cifti2')
def test_read_geometry():
    img = ci.Cifti2Image.from_filename(DATA_FILE6)
    geometry_mapping = img.header.matrix.get_index_map(1)

    # For every brain model in ones.dscalar.nii defines:
    # brain structure name, number of grayordinates, first vertex or voxel, last vertex or voxel
    expected_geometry = [
        ('CIFTI_STRUCTURE_CORTEX_LEFT', 29696, 0, 32491),
        ('CIFTI_STRUCTURE_CORTEX_RIGHT', 29716, 0, 32491),
        ('CIFTI_STRUCTURE_ACCUMBENS_LEFT', 135, [49, 66, 28], [48, 72, 35]),
        ('CIFTI_STRUCTURE_ACCUMBENS_RIGHT', 140, [40, 66, 29], [43, 66, 36]),
        ('CIFTI_STRUCTURE_AMYGDALA_LEFT', 315, [55, 61, 21], [56, 58, 31]),
        ('CIFTI_STRUCTURE_AMYGDALA_RIGHT', 332, [34, 62, 20], [36, 61, 31]),
        ('CIFTI_STRUCTURE_BRAIN_STEM', 3472, [42, 41, 0], [46, 50, 36]),
        ('CIFTI_STRUCTURE_CAUDATE_LEFT', 728, [50, 72, 32], [53, 60, 49]),
        ('CIFTI_STRUCTURE_CAUDATE_RIGHT', 755, [40, 68, 33], [37, 62, 49]),
        ('CIFTI_STRUCTURE_CEREBELLUM_LEFT', 8709, [49, 35, 4], [46, 37, 37]),
        ('CIFTI_STRUCTURE_CEREBELLUM_RIGHT', 9144, [38, 35, 4], [44, 38, 36]),
        ('CIFTI_STRUCTURE_DIENCEPHALON_VENTRAL_LEFT', 706, [52, 53, 26], [56, 49, 35]),
        ('CIFTI_STRUCTURE_DIENCEPHALON_VENTRAL_RIGHT', 712, [39, 54, 26], [35, 49, 36]),
        ('CIFTI_STRUCTURE_HIPPOCAMPUS_LEFT', 764, [55, 60, 21], [54, 44, 39]),
        ('CIFTI_STRUCTURE_HIPPOCAMPUS_RIGHT', 795, [33, 60, 21], [38, 45, 39]),
        ('CIFTI_STRUCTURE_PALLIDUM_LEFT', 297, [56, 59, 32], [55, 61, 39]),
        ('CIFTI_STRUCTURE_PALLIDUM_RIGHT', 260, [36, 62, 32], [35, 62, 39]),
        ('CIFTI_STRUCTURE_PUTAMEN_LEFT', 1060, [51, 66, 28], [58, 64, 43]),
        ('CIFTI_STRUCTURE_PUTAMEN_RIGHT', 1010, [34, 66, 29], [31, 62, 43]),
        ('CIFTI_STRUCTURE_THALAMUS_LEFT', 1288, [55, 47, 33], [52, 53, 46]),
        ('CIFTI_STRUCTURE_THALAMUS_RIGHT', 1248, [32, 47, 34], [38, 55, 46]),
    ]
    current_index = 0
    for from_file, expected in zip(geometry_mapping.brain_models, expected_geometry):
        assert from_file.model_type in ('CIFTI_MODEL_TYPE_SURFACE', 'CIFTI_MODEL_TYPE_VOXELS')
        assert from_file.brain_structure == expected[0]
        assert from_file.index_offset == current_index
        assert from_file.index_count == expected[1]
        current_index += from_file.index_count

        if from_file.model_type == 'CIFTI_MODEL_TYPE_SURFACE':
            assert from_file.voxel_indices_ijk is None
            assert len(from_file.vertex_indices) == expected[1]
            assert from_file.vertex_indices[0] == expected[2]
            assert from_file.vertex_indices[-1] == expected[3]
            assert from_file.surface_number_of_vertices == 32492
        else:
            assert from_file.vertex_indices is None
            assert from_file.surface_number_of_vertices is None
            assert len(from_file.voxel_indices_ijk) == expected[1]
            assert from_file.voxel_indices_ijk[0] == expected[2]
            assert from_file.voxel_indices_ijk[-1] == expected[3]
    assert current_index == img.shape[1]

    expected_affine = [
        [-2, 0, 0, 90],
        [0, 2, 0, -126],
        [0, 0, 2, -72],
        [0, 0, 0, 1],
    ]
    expected_dimensions = (91, 109, 91)
    assert np.array_equal(
        geometry_mapping.volume.transformation_matrix_voxel_indices_ijk_to_xyz.matrix,
        expected_affine,
    )
    assert geometry_mapping.volume.volume_dimensions == expected_dimensions


@needs_nibabel_data('nitest-cifti2')
def test_read_parcels():
    img = ci.Cifti2Image.from_filename(DATA_FILE4)
    parcel_mapping = img.header.matrix.get_index_map(1)

    expected_parcels = [
        ('MEDIAL.WALL', ((719, 20, 28550), (810, 21, 28631))),
        ('BA2_FRB08', ((516, 6757, 17888), (461, 6757, 17887))),
        ('BA1_FRB08', ((211, 5029, 17974), (214, 3433, 17934))),
        ('BA3b_FRB08', ((444, 3436, 18065), (397, 3436, 18065))),
        ('BA4p_FRB08', ((344, 3445, 18164), (371, 3443, 18175))),
        ('BA3a_FRB08', ((290, 3441, 18140), (289, 3440, 18140))),
        ('BA4a_FRB08', ((471, 3446, 18181), (455, 3446, 19759))),
        ('BA6_FRB08', ((1457, 2, 30951), (1400, 2, 30951))),
        ('BA17_V1_FRB08', ((629, 23155, 25785), (635, 23155, 25759))),
        ('BA45_FRB08', ((245, 10100, 18774), (214, 10103, 18907))),
        ('BA44_FRB08', ((226, 10118, 19240), (273, 10119, 19270))),
        ('hOc5_MT_FRB08', ((104, 15019, 23329), (80, 15023, 23376))),
        ('BA18_V2_FRB08', ((702, 95, 25902), (651, 98, 25903))),
        ('V3A_SHM07', ((82, 4, 25050), (82, 4, 25050))),
        ('V3B_SHM07', ((121, 13398, 23303), (121, 13398, 23303))),
        ('LO1_KPO10', ((54, 15007, 23543), (54, 15007, 23543))),
        ('LO2_KPO10', ((79, 15013, 23636), (79, 15013, 23636))),
        ('PITd_KPO10', ((53, 15018, 23769), (65, 15018, 23769))),
        ('PITv_KPO10', ((72, 23480, 23974), (72, 23480, 23974))),
        ('OP1_BSW08', ((470, 8421, 18790), (470, 8421, 18790))),
        ('OP2_BSW08', ((67, 10, 31060), (67, 10, 31060))),
        ('OP3_BSW08', ((119, 10137, 18652), (119, 10137, 18652))),
        ('OP4_BSW08', ((191, 16613, 19429), (192, 16613, 19429))),
        ('IPS1_SHM07', ((54, 11775, 14496), (54, 11775, 14496))),
        ('IPS2_SHM07', ((71, 11771, 14587), (71, 11771, 14587))),
        ('IPS3_SHM07', ((114, 11764, 14783), (114, 11764, 14783))),
        ('IPS4_SHM07', ((101, 11891, 12653), (101, 11891, 12653))),
        ('V7_SHM07', ((140, 11779, 14002), (140, 11779, 14002))),
        ('V4v_SHM07', ((81, 23815, 24557), (90, 23815, 24557))),
        ('V3d_KPO10', ((90, 23143, 25192), (115, 23143, 25192))),
        ('14c_OFP03', ((22, 19851, 21311), (22, 19851, 21311))),
        ('13a_OFP03', ((20, 20963, 21154), (20, 20963, 21154))),
        ('47s_OFP03', ((211, 10182, 20343), (211, 10182, 20343))),
        ('14r_OFP03', ((54, 21187, 21324), (54, 21187, 21324))),
        ('13m_OFP03', ((103, 20721, 21075), (103, 20721, 21075))),
        ('13l_OFP03', ((101, 20466, 20789), (101, 20466, 20789))),
        ('32pl_OFP03', ((14, 19847, 21409), (14, 19847, 21409))),
        ('25_OFP03', ((8, 19844, 27750), (8, 19844, 27750))),
        ('47m_OFP03', ((200, 10174, 20522), (200, 10174, 20522))),
        ('47l_OFP03', ((142, 10164, 19969), (160, 10164, 19969))),
        ('Iai_OFP03', ((153, 10188, 20199), (153, 10188, 20199))),
        ('10r_OFP03', ((138, 19811, 28267), (138, 19811, 28267))),
        ('11m_OFP03', ((92, 20850, 21165), (92, 20850, 21165))),
        ('11l_OFP03', ((200, 20275, 21029), (200, 20275, 21029))),
        ('47r_OFP03', ((259, 10094, 20535), (259, 10094, 20535))),
        ('10m_OFP03', ((102, 19825, 21411), (102, 19825, 21411))),
        ('Iam_OFP03', ((15, 20346, 20608), (15, 20346, 20608))),
        ('Ial_OFP03', ((89, 10194, 11128), (89, 10194, 11128))),
        ('24_OFP03', ((39, 19830, 28279), (36, 19830, 28279))),
        ('Iapm_OFP03', ((7, 20200, 20299), (7, 20200, 20299))),
        ('10p_OFP03', ((480, 19780, 28640), (480, 19780, 28640))),
        ('V6_PHG06', ((72, 12233, 12869), (72, 12233, 12869))),
        ('ER_FRB08', ((103, 21514, 26470), (103, 21514, 26470))),
        ('13b_OFP03', ((60, 21042, 21194), (71, 21040, 21216))),
    ]

    assert img.shape[1] == len(expected_parcels)
    assert len(list(parcel_mapping.parcels)) == len(expected_parcels)

    for (name, expected_surfaces), parcel in zip(expected_parcels, parcel_mapping.parcels):
        assert parcel.name == name
        assert len(parcel.vertices) == 2
        for vertices, orientation, (length, first_element, last_element) in zip(
            parcel.vertices, ('LEFT', 'RIGHT'), expected_surfaces
        ):
            assert len(vertices) == length
            assert vertices[0] == first_element
            assert vertices[-1] == last_element
            assert vertices.brain_structure == f'CIFTI_STRUCTURE_CORTEX_{orientation}'


@needs_nibabel_data('nitest-cifti2')
def test_read_scalar():
    img = ci.Cifti2Image.from_filename(DATA_FILE2)
    scalar_mapping = img.header.matrix.get_index_map(0)

    expected_names = ('MyelinMap_BC_decurv', 'corrThickness')
    assert img.shape[0] == len(expected_names)
    assert len(list(scalar_mapping.named_maps)) == len(expected_names)

    expected_meta = [('PaletteColorMapping', '<PaletteColorMapping Version="1">\n   <ScaleMo')]
    for scalar, name in zip(scalar_mapping.named_maps, expected_names):
        assert scalar.map_name == name

        assert len(scalar.metadata) == len(expected_meta)
        print(expected_meta[0], scalar.metadata.data.keys())
        for key, value in expected_meta:
            assert key in scalar.metadata.data.keys()
            assert scalar.metadata[key][: len(value)] == value

        assert scalar.label_table is None, '.dscalar file should not define a label table'


@needs_nibabel_data('nitest-cifti2')
def test_read_series():
    img = ci.Cifti2Image.from_filename(DATA_FILE4)
    series_mapping = img.header.matrix.get_index_map(0)
    assert series_mapping.series_start == 0.0
    assert series_mapping.series_step == 1.0
    assert series_mapping.series_unit == 'SECOND'
    assert series_mapping.series_exponent == 0.0
    assert series_mapping.number_of_series_points == img.shape[0]


@needs_nibabel_data('nitest-cifti2')
def test_read_labels():
    img = ci.Cifti2Image.from_filename(DATA_FILE5)
    label_mapping = img.header.matrix.get_index_map(0)

    expected_names = [
        'Composite Parcellation-lh (FRB08_OFP03_retinotopic)',
        'Brodmann lh (from colin.R via pals_R-to-fs_LR)',
        'MEDIAL WALL lh (fs_LR)',
    ]
    assert img.shape[0] == len(expected_names)
    assert len(list(label_mapping.named_maps)) == len(expected_names)

    some_expected_labels = {
        0: ('???', (0.667, 0.667, 0.667, 0.0)),
        1: ('MEDIAL.WALL', (0.075, 0.075, 0.075, 1.0)),
        2: ('BA2_FRB08', (0.467, 0.459, 0.055, 1.0)),
        3: ('BA1_FRB08', (0.475, 0.722, 0.859, 1.0)),
        4: ('BA3b_FRB08', (0.855, 0.902, 0.286, 1.0)),
        5: ('BA4p_FRB08', (0.902, 0.573, 0.122, 1.0)),
        89: ('36_B05', (0.467, 0.0, 0.129, 1.0)),
        90: ('35_B05', (0.467, 0.067, 0.067, 1.0)),
        91: ('28_B05', (0.467, 0.337, 0.271, 1.0)),
        92: ('29_B05', (0.267, 0.0, 0.529, 1.0)),
        93: ('26_B05', (0.757, 0.2, 0.227, 1.0)),
        94: ('33_B05', (0.239, 0.082, 0.373, 1.0)),
        95: ('13b_OFP03', (1.0, 1.0, 0.0, 1.0)),
    }

    for named_map, name in zip(label_mapping.named_maps, expected_names):
        assert named_map.map_name == name
        assert len(named_map.metadata) == 0
        assert len(named_map.label_table) == 96
        for index, (label, rgba) in some_expected_labels.items():
            assert named_map.label_table[index].label == label
            assert named_map.label_table[index].rgba == rgba


class TestCifti2SingleHeader(tn2.TestNifti2SingleHeader):
    header_class = _Cifti2AsNiftiHeader
    _pixdim_message = 'pixdim[1,2,3] should be zero or positive'

    def test_pixdim_checks(self):
        hdr_t = self.header_class()
        for i in (1, 2, 3):
            hdr = hdr_t.copy()
            hdr['pixdim'][i] = -1
            assert self._dxer(hdr) == self._pixdim_message

    def test_nifti_qfac_checks(self):
        # Test qfac is 1 or -1 or 0
        hdr = self.header_class()
        # 1, 0, -1 all OK
        hdr['pixdim'][0] = 1
        self.log_chk(hdr, 0)
        hdr['pixdim'][0] = 0
        self.log_chk(hdr, 0)
        hdr['pixdim'][0] = -1
        self.log_chk(hdr, 0)
        # Anything else is not
        hdr['pixdim'][0] = 2
        fhdr, message, raiser = self.log_chk(hdr, 20)
        assert fhdr['pixdim'][0] == 1
        assert message == 'pixdim[0] (qfac) should be 1 (default) or 0 or -1; setting qfac to 1'

    def test_pixdim_log_checks(self):
        # pixdim can be zero or positive
        HC = self.header_class
        hdr = HC()
        hdr['pixdim'][1] = -2  # severity 35
        fhdr, message, raiser = self.log_chk(hdr, 35)
        assert fhdr['pixdim'][1] == 2
        assert message == self._pixdim_message + '; setting to abs of pixdim values'

        pytest.raises(*raiser)

        hdr = HC()
        hdr['pixdim'][1:4] = 0  # No error or warning
        fhdr, message, raiser = self.log_chk(hdr, 0)
        assert raiser == ()



"""Test slicing of file-like objects"""

import time
from functools import partial
from io import BytesIO
from itertools import product
from threading import Lock, Thread

import numpy as np
import pytest
from numpy.testing import assert_array_equal

from ..fileslice import (
    _positive_slice,
    _simple_fileslice,
    calc_slicedefs,
    canonical_slicers,
    fileslice,
    fill_slicer,
    is_fancy,
    optimize_read_slicers,
    optimize_slicer,
    predict_shape,
    read_segments,
    slice2len,
    slice2outax,
    slicers2segments,
    strided_scalar,
    threshold_heuristic,
)


def _check_slice(sliceobj):
    # Fancy indexing always returns a copy, basic indexing returns a view
    a = np.arange(100).reshape((10, 10))
    b = a[sliceobj]
    if np.isscalar(b):
        return  # Can't check
    # Check if this is a view
    a[:] = 99
    b_is_view = np.all(b == 99)
    assert (not is_fancy(sliceobj)) == b_is_view


def test_is_fancy():
    slices = (2, [2], [2, 3], Ellipsis, np.array((2, 3)))
    for slice0 in slices:
        _check_slice(slice0)
        _check_slice((slice0,))  # tuple is same
        # Double ellipsis illegal in np 1.12dev - set up check for that case
        maybe_bad = slice0 is Ellipsis
        for slice1 in slices:
            if maybe_bad and slice1 is Ellipsis:
                continue
            _check_slice((slice0, slice1))
    assert not is_fancy((None,))
    assert not is_fancy((None, 1))
    assert not is_fancy((1, None))
    # Check that actual False returned (rather than falsey)
    assert is_fancy(1) is False


def test_canonical_slicers():
    # Check transformation of sliceobj into canonical form
    slicers = (slice(None), slice(9), slice(0, 9), slice(1, 10), slice(1, 10, 2), 2, np.array(2))

    shape = (10, 10)
    for slice0 in slicers:
        assert canonical_slicers((slice0,), shape) == (slice0, slice(None))
        for slice1 in slicers:
            sliceobj = (slice0, slice1)
            assert canonical_slicers(sliceobj, shape) == sliceobj
            assert canonical_slicers(sliceobj, shape + (2, 3, 4)) == sliceobj + (slice(None),) * 3
            assert canonical_slicers(sliceobj * 3, shape * 3) == sliceobj * 3
            # Check None passes through
            assert canonical_slicers(sliceobj + (None,), shape) == sliceobj + (None,)
            assert canonical_slicers((None,) + sliceobj, shape) == (None,) + sliceobj
            assert canonical_slicers((None,) + sliceobj + (None,), shape) == (None,) + sliceobj + (
                None,
            )
    # Check Ellipsis
    assert canonical_slicers((Ellipsis,), shape) == (slice(None), slice(None))
    assert canonical_slicers((Ellipsis, None), shape) == (slice(None), slice(None), None)
    assert canonical_slicers((Ellipsis, 1), shape) == (slice(None), 1)
    assert canonical_slicers((1, Ellipsis), shape) == (1, slice(None))
    # Ellipsis at end does nothing
    assert canonical_slicers((1, 1, Ellipsis), shape) == (1, 1)
    assert canonical_slicers((1, Ellipsis, 2), (10, 1, 2, 3, 11)) == (
        1,
        slice(None),
        slice(None),
        slice(None),
        2,
    )
    with pytest.raises(ValueError):
        canonical_slicers((Ellipsis, 1, Ellipsis), (2, 3, 4, 5))
    # Check full slices get expanded
    for slice0 in (slice(10), slice(0, 10), slice(0, 10, 1)):
        assert canonical_slicers((slice0, 1), shape) == (slice(None), 1)
    for slice0 in (slice(10), slice(0, 10), slice(0, 10, 1)):
        assert canonical_slicers((slice0, 1), shape) == (slice(None), 1)
        assert canonical_slicers((1, slice0), shape) == (1, slice(None))
    # Check ints etc get parsed through to tuples
    assert canonical_slicers(1, shape) == (1, slice(None))
    assert canonical_slicers(slice(None), shape) == (slice(None), slice(None))
    # Check fancy indexing raises error
    with pytest.raises(ValueError):
        canonical_slicers((np.array([1]), 1), shape)
    with pytest.raises(ValueError):
        canonical_slicers((1, np.array([1])), shape)
    # Check out of range integer raises error
    with pytest.raises(ValueError):
        canonical_slicers((10,), shape)
    with pytest.raises(ValueError):
        canonical_slicers((1, 10), shape)
    with pytest.raises(ValueError):
        canonical_slicers((10,), shape, True)
    with pytest.raises(ValueError):
        canonical_slicers((1, 10), shape, True)
    # Unless check_inds is False
    assert canonical_slicers((10,), shape, False) == (10, slice(None))
    assert canonical_slicers((1, 10), shape, False) == (1, 10)
    # Check negative -> positive
    assert canonical_slicers(-1, shape) == (9, slice(None))
    assert canonical_slicers((slice(None), -1), shape) == (slice(None), 9)
    # check numpy integer scalars behave the same as numpy integers
    assert canonical_slicers(np.array(2), shape) == canonical_slicers(2, shape)
    assert canonical_slicers((np.array(2), np.array(1)), shape) == canonical_slicers((2, 1), shape)
    assert canonical_slicers((2, np.array(1)), shape) == canonical_slicers((2, 1), shape)
    assert canonical_slicers((np.array(2), 1), shape) == canonical_slicers((2, 1), shape)


def test_slice2outax():
    # Test function giving output axes from input ndims and slice
    sn = slice(None)
    assert slice2outax(1, (sn,)) == (0,)
    assert slice2outax(1, (1,)) == (None,)
    assert slice2outax(1, (None,)) == (1,)
    assert slice2outax(1, (None, 1)) == (None,)
    assert slice2outax(1, (None, 1, None)) == (None,)
    assert slice2outax(1, (None, sn)) == (1,)
    assert slice2outax(2, (sn,)) == (0, 1)
    assert slice2outax(2, (sn, sn)) == (0, 1)
    assert slice2outax(2, (1,)) == (None, 0)
    assert slice2outax(2, (sn, 1)) == (0, None)
    assert slice2outax(2, (None,)) == (1, 2)
    assert slice2outax(2, (None, 1)) == (None, 1)
    assert slice2outax(2, (None, 1, None)) == (None, 2)
    assert slice2outax(2, (None, 1, None, 2)) == (None, None)
    assert slice2outax(2, (None, sn, None, 1)) == (1, None)
    assert slice2outax(3, (sn,)) == (0, 1, 2)
    assert slice2outax(3, (sn, sn)) == (0, 1, 2)
    assert slice2outax(3, (sn, None, sn)) == (0, 2, 3)
    assert slice2outax(3, (sn, None, sn, None, sn)) == (0, 2, 4)
    assert slice2outax(3, (1,)) == (None, 0, 1)
    assert slice2outax(3, (None, sn, None, 1)) == (1, None, 3)


def _slices_for_len(L):
    # Example slices for a dimension of length L
    if L == 0:
        raise ValueError('Need length > 0')
    sdefs = [0, L // 2, L - 1, -1, slice(None), slice(L - 1)]
    if L > 1:
        sdefs += [
            -2,
            slice(1, L - 1),
            slice(1, L - 1, 2),
            slice(L - 1, 1, -1),
            slice(L - 1, 1, -2),
        ]
    return tuple(sdefs)


def test_slice2len():
    # Test slice length calculation
    assert slice2len(slice(None), 10) == 10
    assert slice2len(slice(11), 10) == 10
    assert slice2len(slice(1, 11), 10) == 9
    assert slice2len(slice(1, 1), 10) == 0
    assert slice2len(slice(1, 11, 2), 10) == 5
    assert slice2len(slice(0, 11, 3), 10) == 4
    assert slice2len(slice(1, 11, 3), 10) == 3
    assert slice2len(slice(None, None, -1), 10) == 10
    assert slice2len(slice(11, None, -1), 10) == 10
    assert slice2len(slice(None, 1, -1), 10) == 8
    assert slice2len(slice(None, None, -2), 10) == 5
    assert slice2len(slice(None, None, -3), 10) == 4
    assert slice2len(slice(None, 0, -3), 10) == 3
    # Start, end are always taken to be relative if negative
    assert slice2len(slice(None, -4, -1), 10) == 3
    assert slice2len(slice(-4, -2, 1), 10) == 2
    # start after stop
    assert slice2len(slice(3, 2, 1), 10) == 0
    assert slice2len(slice(2, 3, -1), 10) == 0


def test_fill_slicer():
    # Test slice length calculation
    assert fill_slicer(slice(None), 10) == slice(0, 10, 1)
    assert fill_slicer(slice(11), 10) == slice(0, 10, 1)
    assert fill_slicer(slice(1, 11), 10) == slice(1, 10, 1)
    assert fill_slicer(slice(1, 1), 10) == slice(1, 1, 1)
    assert fill_slicer(slice(1, 11, 2), 10) == slice(1, 10, 2)
    assert fill_slicer(slice(0, 11, 3), 10) == slice(0, 10, 3)
    assert fill_slicer(slice(1, 11, 3), 10) == slice(1, 10, 3)
    assert fill_slicer(slice(None, None, -1), 10) == slice(9, None, -1)
    assert fill_slicer(slice(11, None, -1), 10) == slice(9, None, -1)
    assert fill_slicer(slice(None, 1, -1), 10) == slice(9, 1, -1)
    assert fill_slicer(slice(None, None, -2), 10) == slice(9, None, -2)
    assert fill_slicer(slice(None, None, -3), 10) == slice(9, None, -3)
    assert fill_slicer(slice(None, 0, -3), 10) == slice(9, 0, -3)
    # Start, end are always taken to be relative if negative
    assert fill_slicer(slice(None, -4, -1), 10) == slice(9, 6, -1)
    assert fill_slicer(slice(-4, -2, 1), 10) == slice(6, 8, 1)
    # start after stop
    assert fill_slicer(slice(3, 2, 1), 10) == slice(3, 2, 1)
    assert fill_slicer(slice(2, 3, -1), 10) == slice(2, 3, -1)


def test__positive_slice():
    # Reverse slice direction to be positive
    assert _positive_slice(slice(0, 5, 1)) == slice(0, 5, 1)
    assert _positive_slice(slice(1, 5, 3)) == slice(1, 5, 3)
    assert _positive_slice(slice(4, None, -2)) == slice(0, 5, 2)
    assert _positive_slice(slice(4, None, -1)) == slice(0, 5, 1)
    assert _positive_slice(slice(4, 1, -1)) == slice(2, 5, 1)
    assert _positive_slice(slice(4, 1, -2)) == slice(2, 5, 2)


def test_threshold_heuristic():
    # Test for default skip / read heuristic
    # int
    assert threshold_heuristic(1, 9, 1, skip_thresh=8) == 'full'
    assert threshold_heuristic(1, 9, 1, skip_thresh=7) is None
    assert threshold_heuristic(1, 9, 2, skip_thresh=16) == 'full'
    assert threshold_heuristic(1, 9, 2, skip_thresh=15) is None
    # full slice, smallest step size
    assert threshold_heuristic(slice(0, 9, 1), 9, 2, skip_thresh=2) == 'full'
    # Dropping skip thresh below step size gives None
    assert threshold_heuristic(slice(0, 9, 1), 9, 2, skip_thresh=1) == None
    # As does increasing step size
    assert threshold_heuristic(slice(0, 9, 2), 9, 2, skip_thresh=3) == None
    # Negative step size same as positive
    assert threshold_heuristic(slice(9, None, -1), 9, 2, skip_thresh=2) == 'full'
    # Add a gap between start and end. Now contiguous because of step size
    assert threshold_heuristic(slice(2, 9, 1), 9, 2, skip_thresh=2) == 'contiguous'
    # To not-contiguous, even with step size 1
    assert threshold_heuristic(slice(2, 9, 1), 9, 2, skip_thresh=1) == None
    # Back to full when skip covers gap
    assert threshold_heuristic(slice(2, 9, 1), 9, 2, skip_thresh=4) == 'full'
    # Until it doesn't cover the gap
    assert threshold_heuristic(slice(2, 9, 1), 9, 2, skip_thresh=3) == 'contiguous'


# Some dummy heuristics for optimize_slicer
def _always(slicer, dim_len, stride):
    return 'full'


def _partial(slicer, dim_len, stride):
    return 'contiguous'


def _never(slicer, dim_len, stride):
    return None


def test_optimize_slicer():
    # Analyze slice for fullness, contiguity, direction
    #
    # If all_full:
    # - make positive slicer
    # - decide if worth reading continuous block
    # - if so, modify as_read, as_returned accordingly, set contiguous / full
    # - if not, fill as_read for non-contiguous case
    # If not all_full
    # - make positive slicer
    for all_full in (True, False):
        for heuristic in (_always, _never, _partial):
            for is_slowest in (True, False):
                # following tests not affected by all_full or optimization
                # full - always passes through
                assert optimize_slicer(slice(None), 10, all_full, is_slowest, 4, heuristic) == (
                    slice(None),
                    slice(None),
                )
                # Even if full specified with explicit values
                assert optimize_slicer(slice(10), 10, all_full, is_slowest, 4, heuristic) == (
                    slice(None),
                    slice(None),
                )
                assert optimize_slicer(slice(0, 10), 10, all_full, is_slowest, 4, heuristic) == (
                    slice(None),
                    slice(None),
                )
                assert optimize_slicer(
                    slice(0, 10, 1), 10, all_full, is_slowest, 4, heuristic
                ) == (slice(None), slice(None))
                # Reversed full is still full, but with reversed post_slice
                assert optimize_slicer(
                    slice(None, None, -1), 10, all_full, is_slowest, 4, heuristic
                ) == (slice(None), slice(None, None, -1))
    # Contiguous is contiguous unless heuristic kicks in, in which case it may
    # be 'full'
    assert optimize_slicer(slice(9), 10, False, False, 4, _always) == (slice(0, 9, 1), slice(None))
    assert optimize_slicer(slice(9), 10, True, False, 4, _always) == (slice(None), slice(0, 9, 1))
    # Unless this is the slowest dimension, and all_true is True, in which case
    # we don't update to full
    assert optimize_slicer(slice(9), 10, True, True, 4, _always) == (slice(0, 9, 1), slice(None))
    # Nor if the heuristic won't update
    assert optimize_slicer(slice(9), 10, True, False, 4, _never) == (slice(0, 9, 1), slice(None))
    assert optimize_slicer(slice(1, 10), 10, True, False, 4, _never) == (
        slice(1, 10, 1),
        slice(None),
    )
    # Reversed contiguous still contiguous
    assert optimize_slicer(slice(8, None, -1), 10, False, False, 4, _never) == (
        slice(0, 9, 1),
        slice(None, None, -1),
    )
    assert optimize_slicer(slice(8, None, -1), 10, True, False, 4, _always) == (
        slice(None),
        slice(8, None, -1),
    )
    assert optimize_slicer(slice(8, None, -1), 10, False, False, 4, _never) == (
        slice(0, 9, 1),
        slice(None, None, -1),
    )
    assert optimize_slicer(slice(9, 0, -1), 10, False, False, 4, _never) == (
        slice(1, 10, 1),
        slice(None, None, -1),
    )
    # Non-contiguous
    assert optimize_slicer(slice(0, 10, 2), 10, False, False, 4, _never) == (
        slice(0, 10, 2),
        slice(None),
    )
    # all_full triggers optimization, but optimization does nothing
    assert optimize_slicer(slice(0, 10, 2), 10, True, False, 4, _never) == (
        slice(0, 10, 2),
        slice(None),
    )
    # all_full triggers optimization, optimization does something
    assert optimize_slicer(slice(0, 10, 2), 10, True, False, 4, _always) == (
        slice(None),
        slice(0, 10, 2),
    )
    # all_full disables optimization, optimization does something
    assert optimize_slicer(slice(0, 10, 2), 10, False, False, 4, _always) == (
        slice(0, 10, 2),
        slice(None),
    )
    # Non contiguous, reversed
    assert optimize_slicer(slice(10, None, -2), 10, False, False, 4, _never) == (
        slice(1, 10, 2),
        slice(None, None, -1),
    )
    assert optimize_slicer(slice(10, None, -2), 10, True, False, 4, _always) == (
        slice(None),
        slice(9, None, -2),
    )
    # Short non-contiguous
    assert optimize_slicer(slice(2, 8, 2), 10, False, False, 4, _never) == (
        slice(2, 8, 2),
        slice(None),
    )
    # with partial read
    assert optimize_slicer(slice(2, 8, 2), 10, True, False, 4, _partial) == (
        slice(2, 8, 1),
        slice(None, None, 2),
    )
    # If this is the slowest changing dimension, heuristic can upgrade None to
    # contiguous, but not (None, contiguous) to full
    # we've done this one already
    assert optimize_slicer(slice(0, 10, 2), 10, True, False, 4, _always) == (
        slice(None),
        slice(0, 10, 2),
    )
    # if slowest, just upgrade to contiguous
    assert optimize_slicer(slice(0, 10, 2), 10, True, True, 4, _always) == (
        slice(0, 10, 1),
        slice(None, None, 2),
    )
    # contiguous does not upgrade to full
    assert optimize_slicer(slice(9), 10, True, True, 4, _always) == (slice(0, 9, 1), slice(None))
    # integer
    assert optimize_slicer(0, 10, True, False, 4, _never) == (0, 'dropped')
    # can be negative
    assert optimize_slicer(-1, 10, True, False, 4, _never) == (9, 'dropped')
    # or float
    assert optimize_slicer(0.9, 10, True, False, 4, _never) == (0, 'dropped')
    # should never get 'contiguous'
    with pytest.raises(ValueError):
        optimize_slicer(0, 10, True, False, 4, _partial)
    # full can be forced with heuristic
    assert optimize_slicer(0, 10, True, False, 4, _always) == (slice(None), 0)
    # but disabled for slowest changing dimension
    assert optimize_slicer(0, 10, True, True, 4, _always) == (0, 'dropped')


def test_optimize_read_slicers():
    # Test function to optimize read slicers
    assert optimize_read_slicers((1,), (10,), 4, _never) == ((1,), ())
    assert optimize_read_slicers((slice(None),), (10,), 4, _never) == (
        (slice(None),),
        (slice(None),),
    )
    assert optimize_read_slicers((slice(9),), (10,), 4, _never) == (
        (slice(0, 9, 1),),
        (slice(None),),
    )
    # optimize cannot update a continuous to a full if last
    assert optimize_read_slicers((slice(9),), (10,), 4, _always) == (
        (slice(0, 9, 1),),
        (slice(None),),
    )
    # optimize can update non-contiguous to continuous even if last
    # not optimizing
    assert optimize_read_slicers((slice(0, 9, 2),), (10,), 4, _never) == (
        (slice(0, 9, 2),),
        (slice(None),),
    )
    # optimizing
    assert optimize_read_slicers((slice(0, 9, 2),), (10,), 4, _always) == (
        (slice(0, 9, 1),),
        (slice(None, None, 2),),
    )
    # Optimize does nothing for integer when last
    assert optimize_read_slicers((1,), (10,), 4, _always) == ((1,), ())
    # 2D
    assert optimize_read_slicers((slice(None), slice(None)), (10, 6), 4, _never) == (
        (slice(None), slice(None)),
        (slice(None), slice(None)),
    )
    assert optimize_read_slicers((slice(None), 1), (10, 6), 4, _never) == (
        (slice(None), 1),
        (slice(None),),
    )
    assert optimize_read_slicers((1, slice(None)), (10, 6), 4, _never) == (
        (1, slice(None)),
        (slice(None),),
    )
    # Not optimizing a partial slice
    assert optimize_read_slicers((slice(9), slice(None)), (10, 6), 4, _never) == (
        (slice(0, 9, 1), slice(None)),
        (slice(None), slice(None)),
    )
    # Optimizing a partial slice
    assert optimize_read_slicers((slice(9), slice(None)), (10, 6), 4, _always) == (
        (slice(None), slice(None)),
        (slice(0, 9, 1), slice(None)),
    )
    # Optimize cannot update a continuous to a full if last
    assert optimize_read_slicers((slice(None), slice(5)), (10, 6), 4, _always) == (
        (slice(None), slice(0, 5, 1)),
        (slice(None), slice(None)),
    )
    # optimize can update non-contiguous to full if not last
    # not optimizing
    assert optimize_read_slicers((slice(0, 9, 3), slice(None)), (10, 6), 4, _never) == (
        (slice(0, 9, 3), slice(None)),
        (slice(None), slice(None)),
    )
    # optimizing full
    assert optimize_read_slicers((slice(0, 9, 3), slice(None)), (10, 6), 4, _always) == (
        (slice(None), slice(None)),
        (slice(0, 9, 3), slice(None)),
    )
    # optimizing partial
    assert optimize_read_slicers((slice(0, 9, 3), slice(None)), (10, 6), 4, _partial) == (
        (slice(0, 9, 1), slice(None)),
        (slice(None, None, 3), slice(None)),
    )
    # optimize can update non-contiguous to continuous even if last
    # not optimizing
    assert optimize_read_slicers((slice(None), slice(0, 5, 2)), (10, 6), 4, _never) == (
        (slice(None), slice(0, 5, 2)),
        (slice(None), slice(None)),
    )
    # optimizing
    assert optimize_read_slicers((slice(None), slice(0, 5, 2)), (10, 6), 4, _always) == (
        (slice(None), slice(0, 5, 1)),
        (slice(None), slice(None, None, 2)),
    )
    # Optimize does nothing for integer when last
    assert optimize_read_slicers((slice(None), 1), (10, 6), 4, _always) == (
        (slice(None), 1),
        (slice(None),),
    )
    # Check gap threshold with 3D
    depends0 = partial(threshold_heuristic, skip_thresh=10 * 4 - 1)
    depends1 = partial(threshold_heuristic, skip_thresh=10 * 4)
    assert optimize_read_slicers(
        (slice(9), slice(None), slice(None)), (10, 6, 2), 4, depends0
    ) == ((slice(None), slice(None), slice(None)), (slice(0, 9, 1), slice(None), slice(None)))
    assert optimize_read_slicers(
        (slice(None), slice(5), slice(None)), (10, 6, 2), 4, depends0
    ) == ((slice(None), slice(0, 5, 1), slice(None)), (slice(None), slice(None), slice(None)))
    assert optimize_read_slicers(
        (slice(None), slice(5), slice(None)), (10, 6, 2), 4, depends1
    ) == ((slice(None), slice(None), slice(None)), (slice(None), slice(0, 5, 1), slice(None)))
    # Check longs as integer slices
    sn = slice(None)
    assert optimize_read_slicers((1, 2, 3), (2, 3, 4), 4, _always) == ((sn, sn, 3), (1, 2))


def test_slicers2segments():
    # Test function to construct segments from slice objects
    assert slicers2segments((0,), (10,), 7, 4) == [[7, 4]]
    assert slicers2segments((0, 1), (10, 6), 7, 4) == [[7 + 10 * 4, 4]]
    assert slicers2segments((0, 1, 2), (10, 6, 4), 7, 4) == [[7 + 10 * 4 + 10 * 6 * 2 * 4, 4]]
    assert slicers2segments((slice(None),), (10,), 7, 4) == [[7, 10 * 4]]
    assert slicers2segments((0, slice(None)), (10, 6), 7, 4) == [
        [7 + 10 * 4 * i, 4] for i in range(6)
    ]
    assert slicers2segments((slice(None), 0), (10, 6), 7, 4) == [[7, 10 * 4]]
    assert slicers2segments((slice(None), slice(None)), (10, 6), 7, 4) == [[7, 10 * 6 * 4]]
    assert slicers2segments((slice(None), slice(None), 2), (10, 6, 4), 7, 4) == [
        [7 + 10 * 6 * 2 * 4, 10 * 6 * 4]
    ]


def test_calc_slicedefs():
    # Check get_segments routine.  The tests aren't well organized because I
    # wrote them after the code.  We live and (fail to) learn
    segments, out_shape, new_slicing = calc_slicedefs((1,), (10,), 4, 7, 'F', _never)
    assert segments == [[11, 4]]
    assert new_slicing == ()
    assert out_shape == ()
    assert calc_slicedefs((slice(None),), (10,), 4, 7, 'F', _never) == (
        [[7, 40]],
        (10,),
        (),
    )
    assert calc_slicedefs((slice(9),), (10,), 4, 7, 'F', _never) == (
        [[7, 36]],
        (9,),
        (),
    )
    assert calc_slicedefs((slice(1, 9),), (10,), 4, 7, 'F', _never) == (
        [[11, 32]],
        (8,),
        (),
    )
    # Two dimensions, single slice
    assert calc_slicedefs((0,), (10, 6), 4, 7, 'F', _never) == (
        [[7, 4], [47, 4], [87, 4], [127, 4], [167, 4], [207, 4]],
        (6,),
        (),
    )
    assert calc_slicedefs((0,), (10, 6), 4, 7, 'C', _never) == (
        [[7, 6 * 4]],
        (6,),
        (),
    )
    # Two dimensions, contiguous not full
    assert calc_slicedefs((1, slice(1, 5)), (10, 6), 4, 7, 'F', _never) == (
        [[51, 4], [91, 4], [131, 4], [171, 4]],
        (4,),
        (),
    )
    assert calc_slicedefs((1, slice(1, 5)), (10, 6), 4, 7, 'C', _never) == (
        [[7 + 7 * 4, 16]],
        (4,),
        (),
    )
    # With full slice first
    assert calc_slicedefs((slice(None), slice(1, 5)), (10, 6), 4, 7, 'F', _never) == (
        [[47, 160]],
        (10, 4),
        (),
    )
    # Check effect of heuristic on calc_slicedefs
    # Even integer slices can generate full when heuristic says so
    assert calc_slicedefs((1, slice(None)), (10, 6), 4, 7, 'F', _always) == (
        [[7, 10 * 6 * 4]],
        (10, 6),
        (1, slice(None)),
    )
    # Except when last
    assert calc_slicedefs((slice(None), 1), (10, 6), 4, 7, 'F', _always) == (
        [[7 + 10 * 4, 10 * 4]],
        (10,),
        (),
    )


def test_predict_shape():
    shapes = (15, 16, 17, 18)
    for n_dim in range(len(shapes)):
        shape = shapes[: n_dim + 1]
        arr = np.arange(np.prod(shape)).reshape(shape)
        slicers_list = []
        for i in range(n_dim):
            slicers_list.append(_slices_for_len(shape[i]))
            for sliceobj in product(*slicers_list):
                assert predict_shape(sliceobj, shape) == arr[sliceobj].shape
    # Try some Nones and ellipses
    assert predict_shape((Ellipsis,), (2, 3)) == (2, 3)
    assert predict_shape((Ellipsis, 1), (2, 3)) == (2,)
    assert predict_shape((1, Ellipsis), (2, 3)) == (3,)
    assert predict_shape((1, slice(None), Ellipsis), (2, 3)) == (3,)
    assert predict_shape((None,), (2, 3)) == (1, 2, 3)
    assert predict_shape((None, 1), (2, 3)) == (1, 3)
    assert predict_shape((1, None, slice(None)), (2, 3)) == (1, 3)
    assert predict_shape((1, slice(None), None), (2, 3)) == (3, 1)


def test_strided_scalar():
    # Utility to make numpy array of given shape from scalar using striding
    for shape, scalar in product(
        ((2,), (2, 3), (2, 3, 4)),
        (1, 2, np.int16(3)),
    ):
        expected = np.zeros(shape, dtype=np.array(scalar).dtype) + scalar
        observed = strided_scalar(shape, scalar)
        assert_array_equal(observed, expected)
        assert observed.shape == shape
        assert observed.dtype == expected.dtype
        assert_array_equal(observed.strides, 0)
        # Strided scalars are set as not writeable
        # This addresses a numpy 1.10 breakage of broadcasting a strided
        # array without resizing (see GitHub PR #358)
        assert not observed.flags.writeable

        def setval(x):
            x[..., 0] = 99

        # RuntimeError for numpy < 1.10
        with pytest.raises((RuntimeError, ValueError)):
            setval(observed)
    # Default scalar value is 0
    assert_array_equal(strided_scalar((2, 3, 4)), np.zeros((2, 3, 4)))


def _check_bytes(bytes, arr):
    barr = np.ndarray(arr.shape, arr.dtype, buffer=bytes)
    assert_array_equal(barr, arr)


def test_read_segments():
    # Test segment reading
    fobj = BytesIO()
    arr = np.arange(100, dtype=np.int16)
    fobj.write(arr.tobytes())
    _check_bytes(read_segments(fobj, [(0, 200)], 200), arr)
    _check_bytes(read_segments(fobj, [(0, 100), (100, 100)], 200), arr)
    _check_bytes(read_segments(fobj, [(0, 50), (100, 50)], 100), np.r_[arr[:25], arr[50:75]])
    _check_bytes(read_segments(fobj, [(10, 40), (100, 50)], 90), np.r_[arr[5:25], arr[50:75]])
    _check_bytes(read_segments(fobj, [], 0), arr[0:0])
    # Error conditions
    with pytest.raises(ValueError):
        read_segments(fobj, [], 1)
    with pytest.raises(ValueError):
        read_segments(fobj, [(0, 200)], 199)
    with pytest.raises(Exception):
        read_segments(fobj, [(0, 100), (100, 200)], 199)


def test_read_segments_lock():
    # Test read_segment locking with multiple threads
    fobj = BytesIO()
    arr = np.array(np.random.randint(0, 256, 1000), dtype=np.uint8)
    fobj.write(arr.tobytes())

    # Encourage the interpreter to switch threads between a seek/read pair
    def yielding_read(*args, **kwargs):
        time.sleep(0.001)
        return fobj._real_read(*args, **kwargs)

    fobj._real_read = fobj.read
    fobj.read = yielding_read

    # Generate some random array segments to read from the file
    def random_segments(nsegs):
        segs = []
        nbytes = 0

        for i in range(nsegs):
            seglo = np.random.randint(0, 998)
            seghi = np.random.randint(seglo + 1, 1000)
            seglen = seghi - seglo
            nbytes += seglen
            segs.append([seglo, seglen])

        return segs, nbytes

    # Get the data that should be returned for the given segments
    def get_expected(segs):
        segs = [arr[off : off + length] for off, length in segs]
        return np.concatenate(segs)

    # Read from the file, check the result. We do this task simultaneously in
    # many threads. Each thread that passes adds 1 to numpassed[0]
    numpassed = [0]
    lock = Lock()

    def runtest():
        seg, nbytes = random_segments(1)
        expected = get_expected(seg)
        _check_bytes(read_segments(fobj, seg, nbytes, lock), expected)

        seg, nbytes = random_segments(10)
        expected = get_expected(seg)
        _check_bytes(read_segments(fobj, seg, nbytes, lock), expected)

        with lock:
            numpassed[0] += 1

    threads = [Thread(target=runtest) for i in range(100)]
    [t.start() for t in threads]
    [t.join() for t in threads]
    assert numpassed[0] == len(threads)


def _check_slicer(sliceobj, arr, fobj, offset, order, heuristic=threshold_heuristic):
    new_slice = fileslice(fobj, sliceobj, arr.shape, arr.dtype, offset, order, heuristic)
    assert_array_equal(arr[sliceobj], new_slice)


def slicer_samples(shape):
    """Generator returns slice samples for given `shape`"""
    ndim = len(shape)
    slicers_list = []
    for i in range(ndim):
        slicers_list.append(_slices_for_len(shape[i]))
        yield from product(*slicers_list)
    # Nones and ellipses
    yield (None,)
    if ndim == 0:
        return
    yield (None, 0)
    yield (None, np.array(0))
    yield (0, None)
    yield (np.array(0), None)
    yield (Ellipsis, -1)
    yield (Ellipsis, np.array(-1))
    yield (-1, Ellipsis)
    yield (np.array(-1), Ellipsis)
    yield (None, Ellipsis)
    yield (Ellipsis, None)
    yield (Ellipsis, None, None)
    if ndim == 1:
        return
    yield (0, None, slice(None))
    yield (np.array(0), None, slice(None))
    yield (Ellipsis, -1, None)
    yield (Ellipsis, np.array(-1), None)
    yield (0, Ellipsis, None)
    yield (np.array(0), Ellipsis, None)
    if ndim == 2:
        return
    yield (slice(None), 0, -1, None)
    yield (slice(None), np.array(0), np.array(-1), None)
    yield (np.array(0), slice(None), np.array(-1), None)


def test_fileslice():
    shapes = (15, 16, 17)
    for n_dim in range(1, len(shapes) + 1):
        shape = shapes[:n_dim]
        arr = np.arange(np.prod(shape)).reshape(shape)
        for order in 'FC':
            for offset in (0, 20):
                fobj = BytesIO()
                fobj.write(b'\0' * offset)
                fobj.write(arr.tobytes(order=order))
                for sliceobj in slicer_samples(shape):
                    _check_slicer(sliceobj, arr, fobj, offset, order)


def test_fileslice_dtype():
    # Test that any valid dtype specifier works for fileslice
    sliceobj = (slice(None), slice(2))
    for dt in (np.dtype('int32'), np.int32, 'i4', 'int32', '>i4', '<i4'):
        arr = np.arange(24, dtype=dt).reshape((2, 3, 4))
        fobj = BytesIO(arr.tobytes())
        new_slice = fileslice(fobj, sliceobj, arr.shape, dt)
        assert_array_equal(arr[sliceobj], new_slice)


def test_fileslice_errors():
    # Test fileslice causes error on fancy indexing
    arr = np.arange(24).reshape((2, 3, 4))
    fobj = BytesIO(arr.tobytes())
    _check_slicer((1,), arr, fobj, 0, 'C')
    # Fancy indexing raises error
    with pytest.raises(ValueError):
        fileslice(fobj, (np.array([1]),), (2, 3, 4), arr.dtype)


def test_fileslice_heuristic():
    # Just check that any of several heuristics gives the right answer
    shape = (15, 16, 17)
    arr = np.arange(np.prod(shape)).reshape(shape)
    for heuristic in (_always, _never, _partial, threshold_heuristic):
        for order in 'FC':
            fobj = BytesIO()
            fobj.write(arr.tobytes(order=order))
            sliceobj = (1, slice(0, 15, 2), slice(None))
            _check_slicer(sliceobj, arr, fobj, 0, order, heuristic)
            # Check _simple_fileslice while we're at it - si como no?
            new_slice = _simple_fileslice(
                fobj, sliceobj, arr.shape, arr.dtype, 0, order, heuristic
            )
            assert_array_equal(arr[sliceobj], new_slice)



"""
Convert tractograms (TRK -> TCK).
"""

import argparse
import os

import nibabel as nib


def parse_args():
    DESCRIPTION = 'Convert tractograms (TRK -> TCK).'
    parser = argparse.ArgumentParser(description=DESCRIPTION)
    parser.add_argument(
        'tractograms', metavar='tractogram', nargs='+', help='list of tractograms (.trk).'
    )
    parser.add_argument(
        '-f', '--force', action='store_true', help='overwrite existing output files.'
    )

    args = parser.parse_args()
    return args, parser


def main():
    args, parser = parse_args()
    for tractogram in args.tractograms:
        tractogram_format = nib.streamlines.detect_format(tractogram)
        if tractogram_format is not nib.streamlines.TrkFile:
            print(f"Skipping non TRK file: '{tractogram}'")
            continue

        filename, _ = os.path.splitext(tractogram)
        output_filename = filename + '.tck'
        if os.path.isfile(output_filename) and not args.force:
            print(f"Skipping existing file: '{output_filename}'. Use -f to overwrite.")
            continue

        trk = nib.streamlines.load(tractogram)
        nib.streamlines.save(trk.tractogram, output_filename)



#!python
# emacs: -*- mode: python-mode; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the NiBabel package for the
#   copyright and license terms.
#
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##

import unittest

import pytest

import nibabel as nib
from nibabel.cmdline.conform import main
from nibabel.optpkg import optional_package
from nibabel.testing import get_test_data

_, have_scipy, _ = optional_package('scipy.ndimage')
needs_scipy = unittest.skipUnless(have_scipy, 'These tests need scipy')


@needs_scipy
def test_default(tmpdir):
    infile = get_test_data(fname='anatomical.nii')
    outfile = tmpdir / 'output.nii.gz'
    main([str(infile), str(outfile)])
    assert outfile.isfile()
    c = nib.load(outfile)
    assert c.shape == (256, 256, 256)
    assert c.header.get_zooms() == (1, 1, 1)
    assert nib.orientations.aff2axcodes(c.affine) == ('R', 'A', 'S')

    with pytest.raises(FileExistsError):
        main([str(infile), str(outfile)])

    main([str(infile), str(outfile), '--force'])
    assert outfile.isfile()


@needs_scipy
def test_nondefault(tmpdir):
    infile = get_test_data(fname='anatomical.nii')
    outfile = tmpdir / 'output.nii.gz'
    out_shape = (100, 100, 150)
    voxel_size = (1, 2, 4)
    orientation = 'LAS'
    args = (
        f'{infile} {outfile} --out-shape {" ".join(map(str, out_shape))} '
        f'--voxel-size {" ".join(map(str, voxel_size))} --orientation {orientation}'
    )
    main(args.split())
    assert outfile.isfile()
    c = nib.load(outfile)
    assert c.shape == out_shape
    assert c.header.get_zooms() == voxel_size
    assert nib.orientations.aff2axcodes(c.affine) == tuple(orientation)



""" Tests for netcdf """

import os
from os.path import join as pjoin, dirname
from io import BytesIO
from glob import glob
from contextlib import contextmanager

import numpy as np

import pytest

from ..netcdf import netcdf_file

TEST_DATA_PATH = pjoin(dirname(__file__), 'data')

N_EG_ELS = 11  # number of elements for example variable
VARTYPE_EG = 'b'  # var type for example variable


@contextmanager
def make_simple(*args, **kwargs):
    f = netcdf_file(*args, **kwargs)
    f.history = 'Created for a test'
    f.createDimension('time', N_EG_ELS)
    time = f.createVariable('time', VARTYPE_EG, ('time',))
    time[:] = np.arange(N_EG_ELS)
    time.units = 'days since 2008-01-01'
    f.flush()
    yield f
    f.close()


def assert_simple_truths(ncfileobj):
    assert ncfileobj.history == b'Created for a test'
    time = ncfileobj.variables['time']
    assert time.units == b'days since 2008-01-01'
    assert time.shape == (N_EG_ELS,)
    assert time[-1] == N_EG_ELS - 1


def test_read_write_files(tmp_path):
    fname = str(tmp_path / 'simple.nc')

    with make_simple(fname, 'w') as f:
        pass
    # To read the NetCDF file we just created::
    with netcdf_file(fname) as f:
        # Using mmap is the default
        assert f.use_mmap
        assert_simple_truths(f)

    # Now without mmap
    with netcdf_file(fname, mmap=False) as f:
        # Using mmap is the default
        assert not f.use_mmap
        assert_simple_truths(f)

    # To read the NetCDF file we just created, as file object, no
    # mmap.  When n * n_bytes(var_type) is not divisible by 4, this
    # raised an error in pupynere 1.0.12 and scipy rev 5893, because
    # calculated vsize was rounding up in units of 4 - see
    # https://www.unidata.ucar.edu/software/netcdf/docs/netcdf.html
    fobj = open(fname, 'rb')
    with netcdf_file(fobj) as f:
        # by default, don't use mmap for file-like
        assert not f.use_mmap
        assert_simple_truths(f)


def test_read_write_sio():
    eg_sio1 = BytesIO()
    with make_simple(eg_sio1, 'w') as f1:
        str_val = eg_sio1.getvalue()

    eg_sio2 = BytesIO(str_val)
    with netcdf_file(eg_sio2) as f2:
        assert_simple_truths(f2)

    # Test that error is raised if attempting mmap for sio
    eg_sio3 = BytesIO(str_val)
    with pytest.raises(ValueError):
        netcdf_file(eg_sio3, 'r', True)
    # Test 64-bit offset write / read
    eg_sio_64 = BytesIO()
    with make_simple(eg_sio_64, 'w', version=2) as f_64:
        str_val = eg_sio_64.getvalue()

    eg_sio_64 = BytesIO(str_val)
    with netcdf_file(eg_sio_64) as f_64:
        assert_simple_truths(f_64)
        assert f_64.version_byte == 2
    # also when version 2 explicitly specified
    eg_sio_64 = BytesIO(str_val)
    with netcdf_file(eg_sio_64, version=2) as f_64:
        assert_simple_truths(f_64)
        assert f_64.version_byte == 2


def test_read_example_data():
    # read any example data files
    for fname in glob(pjoin(TEST_DATA_PATH, '*.nc')):
        with netcdf_file(fname, 'r') as f:
            pass
        with netcdf_file(fname, 'r', mmap=False) as f:
            pass


def test_itemset_no_segfault_on_readonly():
    # Regression test for ticket #1202.
    # Open the test file in read-only mode.
    filename = pjoin(TEST_DATA_PATH, 'example_1.nc')
    with netcdf_file(filename, 'r') as f:
        time_var = f.variables['time']

    # time_var.assignValue(42) should raise a RuntimeError--not seg. fault!
    with pytest.raises(RuntimeError):
        time_var.assignValue(42)


def test_write_invalid_dtype():
    dtypes = ['int64', 'uint64']
    if np.dtype('int').itemsize == 8:   # 64-bit machines
        dtypes.append('int')
    if np.dtype('uint').itemsize == 8:   # 64-bit machines
        dtypes.append('uint')

    with netcdf_file(BytesIO(), 'w') as f:
        f.createDimension('time', N_EG_ELS)
        for dt in dtypes:
            with pytest.raises(ValueError):
                f.createVariable('time', dt, ('time',))


def test_flush_rewind():
    stream = BytesIO()
    with make_simple(stream, mode='w') as f:
        x = f.createDimension('x', 4)
        v = f.createVariable('v', 'i2', ['x'])
        v[:] = 1
        f.flush()
        len_single = len(stream.getvalue())
        f.flush()
        len_double = len(stream.getvalue())

    assert len_single == len_double


def test_dtype_specifiers():
    # Numpy 1.7.0-dev had a bug where 'i2' wouldn't work.
    # Specifying np.int16 or similar only works from the same commit as this
    # comment was made.
    with make_simple(BytesIO(), mode='w') as f:
        f.createDimension('x',4)
        f.createVariable('v1', 'i2', ['x'])
        f.createVariable('v2', np.int16, ['x'])
        f.createVariable('v3', np.dtype(np.int16), ['x'])


def test_ticket_1720():
    io = BytesIO()

    items = [0,0.1,0.2,0.3,0.4,0.5,0.6,0.7,0.8,0.9]

    with netcdf_file(io, 'w') as f:
        f.history = 'Created for a test'
        f.createDimension('float_var', 10)
        float_var = f.createVariable('float_var', 'f', ('float_var',))
        float_var[:] = items
        float_var.units = 'metres'
        f.flush()
        contents = io.getvalue()

    io = BytesIO(contents)
    with netcdf_file(io, 'r') as f:
        assert f.history == b'Created for a test'
        float_var = f.variables['float_var']
        assert float_var.units == b'metres'
        assert float_var.shape == (10,)
        assert np.allclose(float_var[:], items)



"""Testing environment settings"""

import os
from os import environ as env
from os.path import abspath
from os.path import join as pjoin

import pytest

from .. import environment as nibe

DATA_KEY = 'NIPY_DATA_PATH'
USER_KEY = 'NIPY_USER_DIR'


@pytest.fixture
def with_environment(request):
    """Setup test environment for some functions that are tested
    in this module. In particular this functions stores attributes
    and other things that we need to stub in some test functions.
    This needs to be done on a function level and not module level because
    each testfunction needs a pristine environment.
    """
    GIVEN_ENV = {}
    GIVEN_ENV['env'] = env.copy()
    yield
    """Restore things that were remembered by the setup_environment function """
    orig_env = GIVEN_ENV['env']
    # Pull keys out into list to avoid altering dictionary during iteration,
    # causing python 3 error
    for key in list(env.keys()):
        if key not in orig_env:
            del env[key]
    env.update(orig_env)


def test_nipy_home():
    # Test logic for nipy home directory
    assert nibe.get_home_dir() == os.path.expanduser('~')


def test_user_dir(with_environment):
    if USER_KEY in env:
        del env[USER_KEY]
    home_dir = nibe.get_home_dir()
    if os.name == 'posix':
        exp = pjoin(home_dir, '.nipy')
    else:
        exp = pjoin(home_dir, '_nipy')
    assert exp == nibe.get_nipy_user_dir()
    env[USER_KEY] = '/a/path'
    assert abspath('/a/path') == nibe.get_nipy_user_dir()


def test_sys_dir():
    sys_dir = nibe.get_nipy_system_dir()
    if os.name == 'nt':
        assert sys_dir == r'C:\etc\nipy'
    elif os.name == 'posix':
        assert sys_dir == r'/etc/nipy'
    else:
        assert sys_dir is None



# emacs: -*- mode: python-mode; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the NiBabel package for the
#   copyright and license terms.
#
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Contexts for *with* statement providing temporary directories"""

import os
import tempfile
from contextlib import contextmanager

try:
    from contextlib import chdir as _chdir
except ImportError:  # PY310

    @contextmanager  # type: ignore[no-redef]
    def _chdir(path):
        cwd = os.getcwd()
        os.chdir(path)
        try:
            yield
        finally:
            os.chdir(cwd)


from .deprecated import deprecate_with_version


class TemporaryDirectory(tempfile.TemporaryDirectory):
    """Create and return a temporary directory.  This has the same
    behavior as mkdtemp but can be used as a context manager.

    Upon exiting the context, the directory and everything contained
    in it are removed.
    """

    @deprecate_with_version(
        'Please use the standard library tempfile.TemporaryDirectory',
        '5.0',
        '7.0',
    )
    def __init__(self, suffix='', prefix=tempfile.template, dir=None):
        """
        Examples
        --------
        >>> import os
        >>> with TemporaryDirectory() as tmpdir:
        ...     fname = os.path.join(tmpdir, 'example_file.txt')
        ...     with open(fname, 'wt') as fobj:
        ...         _ = fobj.write('a string\\n')
        >>> os.path.exists(tmpdir)
        False
        """
        super().__init__(suffix, prefix, dir)


@contextmanager
def InTemporaryDirectory():
    """Create, return, and change directory to a temporary directory

    Notes
    -----
    As its name suggests, the class temporarily changes the working
    directory of the Python process, and this is not thread-safe.  We suggest
    using it only for tests.

    Examples
    --------
    >>> import os
    >>> from pathlib import Path
    >>> my_cwd = os.getcwd()
    >>> with InTemporaryDirectory() as tmpdir:
    ...     _ = Path('test.txt').write_text('some text')
    ...     assert os.path.isfile('test.txt')
    ...     assert os.path.isfile(os.path.join(tmpdir, 'test.txt'))
    >>> os.path.exists(tmpdir)
    False
    >>> os.getcwd() == my_cwd
    True
    """
    with tempfile.TemporaryDirectory() as tmpdir, _chdir(tmpdir):
        yield tmpdir


@contextmanager
def InGivenDirectory(path=None):
    """Change directory to given directory for duration of ``with`` block

    Useful when you want to use `InTemporaryDirectory` for the final test, but
    you are still debugging.  For example, you may want to do this in the end:

    >>> with InTemporaryDirectory() as tmpdir:
    ...     # do something complicated which might break
    ...     pass

    But indeed the complicated thing does break, and meanwhile the
    ``InTemporaryDirectory`` context manager wiped out the directory with the
    temporary files that you wanted for debugging.  So, while debugging, you
    replace with something like:

    >>> with InGivenDirectory() as tmpdir: # Use working directory by default
    ...     # do something complicated which might break
    ...     pass

    You can then look at the temporary file outputs to debug what is happening,
    fix, and finally replace ``InGivenDirectory`` with ``InTemporaryDirectory``
    again.

    Parameters
    ----------
    path : None or str, optional
        path to change directory to, for duration of ``with`` block.
        Defaults to ``os.getcwd()`` if None
    """
    if path is None:
        path = os.getcwd()
    os.makedirs(path, exist_ok=True)
    with _chdir(path):
        yield os.path.abspath(path)



"""Helpers for typing compatibility across Python versions"""

import sys

if sys.version_info < (3, 10):
    from typing_extensions import ParamSpec
else:
    from typing import ParamSpec

if sys.version_info < (3, 11):
    from typing_extensions import Self
else:
    from typing import Self

if sys.version_info < (3, 13):
    from typing_extensions import TypeVar
else:
    from typing import TypeVar


__all__ = [
    'ParamSpec',
    'Self',
    'TypeVar',
]



# emacs: -*- mode: python-mode; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the NiBabel package for the
#   copyright and license terms.
#
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Tests for SPM2 header stuff"""

import numpy as np
import pytest
from numpy.testing import assert_array_equal

from ..spatialimages import HeaderDataError, HeaderTypeError
from ..spm2analyze import Spm2AnalyzeHeader, Spm2AnalyzeImage
from . import test_spm99analyze


class TestSpm2AnalyzeHeader(test_spm99analyze.TestSpm99AnalyzeHeader):
    header_class = Spm2AnalyzeHeader

    def test_slope_inter(self):
        hdr = self.header_class()
        assert hdr.get_slope_inter() == (1.0, 0.0)
        for in_tup, exp_err, out_tup, raw_slope in (
            ((2.0,), None, (2.0, 0.0), 2.0),
            ((None,), None, (None, None), np.nan),
            ((1.0, None), None, (1.0, 0.0), 1.0),
            # non zero intercept causes error
            ((None, 1.1), HeaderTypeError, (None, None), np.nan),
            ((2.0, 1.1), HeaderTypeError, (None, None), 2.0),
            # null scalings
            ((0.0, None), HeaderDataError, (None, None), 0.0),
            ((np.nan, np.nan), None, (None, None), np.nan),
            ((np.nan, None), None, (None, None), np.nan),
            ((None, np.nan), None, (None, None), np.nan),
            ((np.inf, None), HeaderDataError, (None, None), np.inf),
            ((-np.inf, None), HeaderDataError, (None, None), -np.inf),
            ((None, 0.0), None, (None, None), np.nan),
        ):
            hdr = self.header_class()
            if not exp_err is None:
                with pytest.raises(exp_err):
                    hdr.set_slope_inter(*in_tup)
                # raw set
                if not in_tup[0] is None:
                    hdr['scl_slope'] = in_tup[0]
            else:
                hdr.set_slope_inter(*in_tup)
                assert hdr.get_slope_inter() == out_tup
                # Check set survives through checking
                hdr = Spm2AnalyzeHeader.from_header(hdr, check=True)
                assert hdr.get_slope_inter() == out_tup
            assert_array_equal(hdr['scl_slope'], raw_slope)


class TestSpm2AnalyzeImage(test_spm99analyze.TestSpm99AnalyzeImage):
    # class for testing images
    image_class = Spm2AnalyzeImage


def test_origin_affine():
    # check that origin affine works, only
    hdr = Spm2AnalyzeHeader()
    hdr.get_origin_affine()



"""Tests the generation of new CIFTI-2 files from scratch

Contains a series of functions to create and check each of the 5 CIFTI index
types (i.e. BRAIN_MODELS, PARCELS, SCALARS, LABELS, and SERIES).

These functions are used in the tests to generate most CIFTI file types from
scratch.
"""

import numpy as np
import pytest

import nibabel as nib
from nibabel import cifti2 as ci
from nibabel.tmpdirs import InTemporaryDirectory

from ...testing import (
    assert_array_equal,
    clear_and_catch_warnings,
    error_warnings,
    suppress_warnings,
)

affine = [
    [-1.5, 0, 0, 90],
    [0, 1.5, 0, -85],
    [0, 0, 1.5, -71],
    [0, 0, 0, 1.0],
]

dimensions = (120, 83, 78)

number_of_vertices = 30000

brain_models = [
    (
        'CIFTI_STRUCTURE_THALAMUS_LEFT',
        [
            [60, 60, 60],
            [61, 59, 60],
            [61, 60, 59],
            [80, 90, 92],
        ],
    ),
    ('CIFTI_STRUCTURE_CORTEX_LEFT', [0, 1000, 1301, 19972, 27312]),
    ('CIFTI_STRUCTURE_CORTEX_RIGHT', [207]),
]


def create_geometry_map(applies_to_matrix_dimension):
    voxels = ci.Cifti2VoxelIndicesIJK(brain_models[0][1])
    left_thalamus = ci.Cifti2BrainModel(
        index_offset=0,
        index_count=4,
        model_type='CIFTI_MODEL_TYPE_VOXELS',
        brain_structure=brain_models[0][0],
        voxel_indices_ijk=voxels,
    )

    vertices = ci.Cifti2VertexIndices(np.array(brain_models[1][1]))
    left_cortex = ci.Cifti2BrainModel(
        index_offset=4,
        index_count=5,
        model_type='CIFTI_MODEL_TYPE_SURFACE',
        brain_structure=brain_models[1][0],
        vertex_indices=vertices,
    )
    left_cortex.surface_number_of_vertices = number_of_vertices

    vertices = ci.Cifti2VertexIndices(np.array(brain_models[2][1]))
    right_cortex = ci.Cifti2BrainModel(
        index_offset=9,
        index_count=1,
        model_type='CIFTI_MODEL_TYPE_SURFACE',
        brain_structure=brain_models[2][0],
        vertex_indices=vertices,
    )
    right_cortex.surface_number_of_vertices = number_of_vertices

    volume = ci.Cifti2Volume(
        dimensions, ci.Cifti2TransformationMatrixVoxelIndicesIJKtoXYZ(-3, affine)
    )
    return ci.Cifti2MatrixIndicesMap(
        applies_to_matrix_dimension,
        'CIFTI_INDEX_TYPE_BRAIN_MODELS',
        maps=[left_thalamus, left_cortex, right_cortex, volume],
    )


def check_geometry_map(mapping):
    assert mapping.indices_map_to_data_type == 'CIFTI_INDEX_TYPE_BRAIN_MODELS'
    assert len(list(mapping.brain_models)) == 3
    left_thalamus, left_cortex, right_cortex = mapping.brain_models

    assert left_thalamus.index_offset == 0
    assert left_thalamus.index_count == 4
    assert left_thalamus.model_type == 'CIFTI_MODEL_TYPE_VOXELS'
    assert left_thalamus.brain_structure == brain_models[0][0]
    assert left_thalamus.vertex_indices is None
    assert left_thalamus.surface_number_of_vertices is None
    assert left_thalamus.voxel_indices_ijk._indices == brain_models[0][1]

    assert left_cortex.index_offset == 4
    assert left_cortex.index_count == 5
    assert left_cortex.model_type == 'CIFTI_MODEL_TYPE_SURFACE'
    assert left_cortex.brain_structure == brain_models[1][0]
    assert left_cortex.voxel_indices_ijk is None
    assert left_cortex.vertex_indices._indices == brain_models[1][1]
    assert left_cortex.surface_number_of_vertices == number_of_vertices

    assert right_cortex.index_offset == 9
    assert right_cortex.index_count == 1
    assert right_cortex.model_type == 'CIFTI_MODEL_TYPE_SURFACE'
    assert right_cortex.brain_structure == brain_models[2][0]
    assert right_cortex.voxel_indices_ijk is None
    assert right_cortex.vertex_indices._indices == brain_models[2][1]
    assert right_cortex.surface_number_of_vertices == number_of_vertices

    assert mapping.volume.volume_dimensions == dimensions
    assert (mapping.volume.transformation_matrix_voxel_indices_ijk_to_xyz.matrix == affine).all()


parcels = [
    (
        'volume_parcel',
        (
            [
                [60, 60, 60],
                [61, 59, 60],
                [61, 60, 59],
                [80, 90, 92],
            ],
        ),
    ),
    (
        'surface_parcel',
        (
            ('CIFTI_STRUCTURE_CORTEX_LEFT', [0, 1000, 1301, 19972, 27312]),
            ('CIFTI_STRUCTURE_CORTEX_RIGHT', [0, 100, 381]),
        ),
    ),
    (
        'mixed_parcel',
        (
            [
                [71, 81, 39],
                [53, 21, 91],
            ],
            ('CIFTI_STRUCTURE_CORTEX_LEFT', [71, 88, 999]),
        ),
    ),
    ('single_element', ([[71, 81, 39]], ('CIFTI_STRUCTURE_CORTEX_LEFT', [40]))),
]


def create_parcel_map(applies_to_matrix_dimension):
    mapping = ci.Cifti2MatrixIndicesMap(applies_to_matrix_dimension, 'CIFTI_INDEX_TYPE_PARCELS')
    for name, elements in parcels:
        surfaces = []
        volume = None
        for element in elements:
            if isinstance(element[0], str):
                surfaces.append(ci.Cifti2Vertices(element[0], element[1]))
            else:
                volume = ci.Cifti2VoxelIndicesIJK(element)
        mapping.append(ci.Cifti2Parcel(name, volume, surfaces))

    mapping.extend(
        [
            ci.Cifti2Surface(f'CIFTI_STRUCTURE_CORTEX_{orientation}', number_of_vertices)
            for orientation in ['LEFT', 'RIGHT']
        ]
    )
    mapping.volume = ci.Cifti2Volume(
        dimensions, ci.Cifti2TransformationMatrixVoxelIndicesIJKtoXYZ(-3, affine)
    )
    return mapping


def check_parcel_map(mapping):
    assert mapping.indices_map_to_data_type == 'CIFTI_INDEX_TYPE_PARCELS'
    assert len(list(mapping.parcels)) == len(parcels)
    for (name, elements), parcel in zip(parcels, mapping.parcels):
        assert parcel.name == name
        idx_surface = 0
        for element in elements:
            if isinstance(element[0], str):
                surface = parcel.vertices[idx_surface]
                assert surface.brain_structure == element[0]
                assert surface._vertices == element[1]
                idx_surface += 1
            else:
                assert parcel.voxel_indices_ijk._indices == element

    for surface, orientation in zip(mapping.surfaces, ('LEFT', 'RIGHT')):
        assert surface.brain_structure == f'CIFTI_STRUCTURE_CORTEX_{orientation}'
        assert surface.surface_number_of_vertices == number_of_vertices

    assert mapping.volume.volume_dimensions == dimensions
    assert (mapping.volume.transformation_matrix_voxel_indices_ijk_to_xyz.matrix == affine).all()


scalars = [('first_name', {'meta_key': 'some_metadata'}), ('another name', {})]


def create_scalar_map(applies_to_matrix_dimension):
    maps = [ci.Cifti2NamedMap(name, ci.Cifti2MetaData(meta)) for name, meta in scalars]
    return ci.Cifti2MatrixIndicesMap(
        applies_to_matrix_dimension, 'CIFTI_INDEX_TYPE_SCALARS', maps=maps
    )


def check_scalar_map(mapping):
    assert mapping.indices_map_to_data_type == 'CIFTI_INDEX_TYPE_SCALARS'
    assert len(list(mapping.named_maps)) == 2

    for expected, named_map in zip(scalars, mapping.named_maps):
        assert named_map.map_name == expected[0]
        if len(expected[1]) == 0:
            assert named_map.metadata is None
        else:
            assert named_map.metadata == expected[1]


labels = [
    (
        'first_name',
        {'meta_key': 'some_metadata'},
        {
            0: ('label0', (0.1, 0.3, 0.2, 0.5)),
            1: ('new_label', (0.5, 0.3, 0.1, 0.4)),
        },
    ),
    (
        'another name',
        {},
        {
            0: ('???', (0, 0, 0, 0)),
            1: ('great region', (0.4, 0.1, 0.23, 0.15)),
        },
    ),
]


def create_label_map(applies_to_matrix_dimension):
    maps = []
    for name, meta, label in labels:
        label_table = ci.Cifti2LabelTable()
        for key, (tag, rgba) in label.items():
            label_table[key] = ci.Cifti2Label(key, tag, *rgba)
        maps.append(ci.Cifti2NamedMap(name, ci.Cifti2MetaData(meta), label_table))
    return ci.Cifti2MatrixIndicesMap(
        applies_to_matrix_dimension, 'CIFTI_INDEX_TYPE_LABELS', maps=maps
    )


def check_label_map(mapping):
    assert mapping.indices_map_to_data_type == 'CIFTI_INDEX_TYPE_LABELS'
    assert len(list(mapping.named_maps)) == 2

    for expected, named_map in zip(scalars, mapping.named_maps):
        assert named_map.map_name == expected[0]
        if len(expected[1]) == 0:
            assert named_map.metadata is None
        else:
            assert named_map.metadata == expected[1]


def create_series_map(applies_to_matrix_dimension):
    return ci.Cifti2MatrixIndicesMap(
        applies_to_matrix_dimension,
        'CIFTI_INDEX_TYPE_SERIES',
        number_of_series_points=13,
        series_exponent=-3,
        series_start=18.2,
        series_step=10.5,
        series_unit='SECOND',
    )


def check_series_map(mapping):
    assert mapping.indices_map_to_data_type == 'CIFTI_INDEX_TYPE_SERIES'
    assert mapping.number_of_series_points == 13
    assert mapping.series_exponent == -3
    assert mapping.series_start == 18.2
    assert mapping.series_step == 10.5
    assert mapping.series_unit == 'SECOND'


def test_dtseries():
    series_map = create_series_map((0,))
    geometry_map = create_geometry_map((1,))
    matrix = ci.Cifti2Matrix()
    matrix.extend((series_map, geometry_map))
    hdr = ci.Cifti2Header(matrix)
    data = np.random.randn(13, 10)
    img = ci.Cifti2Image(data, hdr)
    img.nifti_header.set_intent('NIFTI_INTENT_CONNECTIVITY_DENSE_SERIES')

    with InTemporaryDirectory():
        ci.save(img, 'test.dtseries.nii')
        img2 = nib.load('test.dtseries.nii')
        assert img2.nifti_header.get_intent()[0] == 'ConnDenseSeries'
        assert isinstance(img2, ci.Cifti2Image)
        assert_array_equal(img2.get_fdata(), data)
        check_series_map(img2.header.matrix.get_index_map(0))
        check_geometry_map(img2.header.matrix.get_index_map(1))
        del img2


def test_dscalar():
    scalar_map = create_scalar_map((0,))
    geometry_map = create_geometry_map((1,))
    matrix = ci.Cifti2Matrix()
    matrix.extend((scalar_map, geometry_map))
    hdr = ci.Cifti2Header(matrix)
    data = np.random.randn(2, 10)
    img = ci.Cifti2Image(data, hdr)
    img.nifti_header.set_intent('NIFTI_INTENT_CONNECTIVITY_DENSE_SCALARS')

    with InTemporaryDirectory():
        ci.save(img, 'test.dscalar.nii')
        img2 = nib.load('test.dscalar.nii')
        assert img2.nifti_header.get_intent()[0] == 'ConnDenseScalar'
        assert isinstance(img2, ci.Cifti2Image)
        assert_array_equal(img2.get_fdata(), data)
        check_scalar_map(img2.header.matrix.get_index_map(0))
        check_geometry_map(img2.header.matrix.get_index_map(1))
        del img2


def test_dlabel():
    label_map = create_label_map((0,))
    geometry_map = create_geometry_map((1,))
    matrix = ci.Cifti2Matrix()
    matrix.extend((label_map, geometry_map))
    hdr = ci.Cifti2Header(matrix)
    data = np.random.randn(2, 10)
    img = ci.Cifti2Image(data, hdr)
    img.nifti_header.set_intent('NIFTI_INTENT_CONNECTIVITY_DENSE_LABELS')

    with InTemporaryDirectory():
        ci.save(img, 'test.dlabel.nii')
        img2 = nib.load('test.dlabel.nii')
        assert img2.nifti_header.get_intent()[0] == 'ConnDenseLabel'
        assert isinstance(img2, ci.Cifti2Image)
        assert_array_equal(img2.get_fdata(), data)
        check_label_map(img2.header.matrix.get_index_map(0))
        check_geometry_map(img2.header.matrix.get_index_map(1))
        del img2


def test_dconn():
    mapping = create_geometry_map((0, 1))
    matrix = ci.Cifti2Matrix()
    matrix.append(mapping)
    hdr = ci.Cifti2Header(matrix)
    data = np.random.randn(10, 10)
    img = ci.Cifti2Image(data, hdr)
    img.nifti_header.set_intent('NIFTI_INTENT_CONNECTIVITY_DENSE')

    with InTemporaryDirectory():
        ci.save(img, 'test.dconn.nii')
        img2 = nib.load('test.dconn.nii')
        assert img2.nifti_header.get_intent()[0] == 'ConnDense'
        assert isinstance(img2, ci.Cifti2Image)
        assert_array_equal(img2.get_fdata(), data)
        assert img2.header.matrix.get_index_map(0) == img2.header.matrix.get_index_map(1)
        check_geometry_map(img2.header.matrix.get_index_map(0))
        del img2


def test_ptseries():
    series_map = create_series_map((0,))
    parcel_map = create_parcel_map((1,))
    matrix = ci.Cifti2Matrix()
    matrix.extend((series_map, parcel_map))
    hdr = ci.Cifti2Header(matrix)
    data = np.random.randn(13, 4)
    img = ci.Cifti2Image(data, hdr)
    img.nifti_header.set_intent('NIFTI_INTENT_CONNECTIVITY_PARCELLATED_SERIES')

    with InTemporaryDirectory():
        ci.save(img, 'test.ptseries.nii')
        img2 = nib.load('test.ptseries.nii')
        assert img2.nifti_header.get_intent()[0] == 'ConnParcelSries'
        assert isinstance(img2, ci.Cifti2Image)
        assert_array_equal(img2.get_fdata(), data)
        check_series_map(img2.header.matrix.get_index_map(0))
        check_parcel_map(img2.header.matrix.get_index_map(1))
        del img2


def test_pscalar():
    scalar_map = create_scalar_map((0,))
    parcel_map = create_parcel_map((1,))
    matrix = ci.Cifti2Matrix()
    matrix.extend((scalar_map, parcel_map))
    hdr = ci.Cifti2Header(matrix)
    data = np.random.randn(2, 4)
    img = ci.Cifti2Image(data, hdr)
    img.nifti_header.set_intent('NIFTI_INTENT_CONNECTIVITY_PARCELLATED_SCALAR')

    with InTemporaryDirectory():
        ci.save(img, 'test.pscalar.nii')
        img2 = nib.load('test.pscalar.nii')
        assert img2.nifti_header.get_intent()[0] == 'ConnParcelScalr'
        assert isinstance(img2, ci.Cifti2Image)
        assert_array_equal(img2.get_fdata(), data)
        check_scalar_map(img2.header.matrix.get_index_map(0))
        check_parcel_map(img2.header.matrix.get_index_map(1))
        del img2


def test_pdconn():
    geometry_map = create_geometry_map((0,))
    parcel_map = create_parcel_map((1,))
    matrix = ci.Cifti2Matrix()
    matrix.extend((geometry_map, parcel_map))
    hdr = ci.Cifti2Header(matrix)
    data = np.random.randn(10, 4)
    img = ci.Cifti2Image(data, hdr)
    img.nifti_header.set_intent('NIFTI_INTENT_CONNECTIVITY_PARCELLATED_DENSE')

    with InTemporaryDirectory():
        ci.save(img, 'test.pdconn.nii')
        img2 = ci.load('test.pdconn.nii')
        assert img2.nifti_header.get_intent()[0] == 'ConnParcelDense'
        assert isinstance(img2, ci.Cifti2Image)
        assert_array_equal(img2.get_fdata(), data)
        check_geometry_map(img2.header.matrix.get_index_map(0))
        check_parcel_map(img2.header.matrix.get_index_map(1))
        del img2


def test_dpconn():
    parcel_map = create_parcel_map((0,))
    geometry_map = create_geometry_map((1,))
    matrix = ci.Cifti2Matrix()
    matrix.extend((parcel_map, geometry_map))
    hdr = ci.Cifti2Header(matrix)
    data = np.random.randn(4, 10)
    img = ci.Cifti2Image(data, hdr)
    img.nifti_header.set_intent('NIFTI_INTENT_CONNECTIVITY_DENSE_PARCELLATED')

    with InTemporaryDirectory():
        ci.save(img, 'test.dpconn.nii')
        img2 = ci.load('test.dpconn.nii')
        assert img2.nifti_header.get_intent()[0] == 'ConnDenseParcel'
        assert isinstance(img2, ci.Cifti2Image)
        assert_array_equal(img2.get_fdata(), data)
        check_parcel_map(img2.header.matrix.get_index_map(0))
        check_geometry_map(img2.header.matrix.get_index_map(1))
        del img2


def test_plabel():
    label_map = create_label_map((0,))
    parcel_map = create_parcel_map((1,))
    matrix = ci.Cifti2Matrix()
    matrix.extend((label_map, parcel_map))
    hdr = ci.Cifti2Header(matrix)
    data = np.random.randn(2, 4)
    img = ci.Cifti2Image(data, hdr)

    with InTemporaryDirectory():
        ci.save(img, 'test.plabel.nii')
        img2 = ci.load('test.plabel.nii')
        assert img.nifti_header.get_intent()[0] == 'ConnUnknown'
        assert isinstance(img2, ci.Cifti2Image)
        assert_array_equal(img2.get_fdata(), data)
        check_label_map(img2.header.matrix.get_index_map(0))
        check_parcel_map(img2.header.matrix.get_index_map(1))
        del img2


def test_pconn():
    mapping = create_parcel_map((0, 1))
    matrix = ci.Cifti2Matrix()
    matrix.append(mapping)
    hdr = ci.Cifti2Header(matrix)
    data = np.random.randn(4, 4)
    img = ci.Cifti2Image(data, hdr)
    img.nifti_header.set_intent('NIFTI_INTENT_CONNECTIVITY_PARCELLATED')

    with InTemporaryDirectory():
        ci.save(img, 'test.pconn.nii')
        img2 = ci.load('test.pconn.nii')
        assert img.nifti_header.get_intent()[0] == 'ConnParcels'
        assert isinstance(img2, ci.Cifti2Image)
        assert_array_equal(img2.get_fdata(), data)
        assert img2.header.matrix.get_index_map(0) == img2.header.matrix.get_index_map(1)
        check_parcel_map(img2.header.matrix.get_index_map(0))
        del img2


def test_pconnseries():
    parcel_map = create_parcel_map((0, 1))
    series_map = create_series_map((2,))

    matrix = ci.Cifti2Matrix()
    matrix.extend((parcel_map, series_map))
    hdr = ci.Cifti2Header(matrix)
    data = np.random.randn(4, 4, 13)
    img = ci.Cifti2Image(data, hdr)
    img.nifti_header.set_intent('NIFTI_INTENT_CONNECTIVITY_PARCELLATED_PARCELLATED_SERIES')

    with InTemporaryDirectory():
        ci.save(img, 'test.pconnseries.nii')
        img2 = ci.load('test.pconnseries.nii')
        assert img.nifti_header.get_intent()[0] == 'ConnPPSr'
        assert isinstance(img2, ci.Cifti2Image)
        assert_array_equal(img2.get_fdata(), data)
        assert img2.header.matrix.get_index_map(0) == img2.header.matrix.get_index_map(1)
        check_parcel_map(img2.header.matrix.get_index_map(0))
        check_series_map(img2.header.matrix.get_index_map(2))
        del img2


def test_pconnscalar():
    parcel_map = create_parcel_map((0, 1))
    scalar_map = create_scalar_map((2,))

    matrix = ci.Cifti2Matrix()
    matrix.extend((parcel_map, scalar_map))
    hdr = ci.Cifti2Header(matrix)
    data = np.random.randn(4, 4, 2)
    img = ci.Cifti2Image(data, hdr)
    img.nifti_header.set_intent('NIFTI_INTENT_CONNECTIVITY_PARCELLATED_PARCELLATED_SCALAR')

    with InTemporaryDirectory():
        ci.save(img, 'test.pconnscalar.nii')
        img2 = ci.load('test.pconnscalar.nii')
        assert img.nifti_header.get_intent()[0] == 'ConnPPSc'
        assert isinstance(img2, ci.Cifti2Image)
        assert_array_equal(img2.get_fdata(), data)
        assert img2.header.matrix.get_index_map(0) == img2.header.matrix.get_index_map(1)

        check_parcel_map(img2.header.matrix.get_index_map(0))
        check_scalar_map(img2.header.matrix.get_index_map(2))
        del img2


def test_wrong_shape():
    scalar_map = create_scalar_map((0,))
    brain_model_map = create_geometry_map((1,))

    matrix = ci.Cifti2Matrix()
    matrix.extend((scalar_map, brain_model_map))
    hdr = ci.Cifti2Header(matrix)

    # correct shape is (2, 10)
    for data in (
        np.random.randn(1, 11),
        np.random.randn(2, 10, 1),
        np.random.randn(1, 2, 10),
        np.random.randn(3, 10),
        np.random.randn(2, 9),
    ):
        with clear_and_catch_warnings():
            with error_warnings():
                with pytest.raises(UserWarning):
                    ci.Cifti2Image(data, hdr)
        with suppress_warnings():
            img = ci.Cifti2Image(data, hdr)

        with pytest.raises(ValueError):
            img.to_file_map()



"""
Convert tractograms (TCK -> TRK).
"""

import argparse
import os

import nibabel as nib
from nibabel.orientations import aff2axcodes
from nibabel.streamlines import Field


def parse_args():
    DESCRIPTION = 'Convert tractograms (TCK -> TRK).'
    parser = argparse.ArgumentParser(description=DESCRIPTION)
    parser.add_argument('anatomy', help='reference anatomical image (.nii|.nii.gz.')
    parser.add_argument(
        'tractograms', metavar='tractogram', nargs='+', help='list of tractograms (.tck).'
    )
    parser.add_argument(
        '-f', '--force', action='store_true', help='overwrite existing output files.'
    )

    args = parser.parse_args()
    return args, parser


def main():
    args, parser = parse_args()

    try:
        nii = nib.load(args.anatomy)
    except Exception:
        parser.error('Expecting anatomical image as first argument.')

    for tractogram in args.tractograms:
        tractogram_format = nib.streamlines.detect_format(tractogram)
        if tractogram_format is not nib.streamlines.TckFile:
            print(f"Skipping non TCK file: '{tractogram}'")
            continue

        filename, _ = os.path.splitext(tractogram)
        output_filename = filename + '.trk'
        if os.path.isfile(output_filename) and not args.force:
            print(f"Skipping existing file: '{output_filename}'. Use -f to overwrite.")
            continue

        # Build header using infos from the anatomical image.
        header = {}
        header[Field.VOXEL_TO_RASMM] = nii.affine.copy()
        header[Field.VOXEL_SIZES] = nii.header.get_zooms()[:3]
        header[Field.DIMENSIONS] = nii.shape[:3]
        header[Field.VOXEL_ORDER] = ''.join(aff2axcodes(nii.affine))

        tck = nib.streamlines.load(tractogram)
        nib.streamlines.save(tck.tractogram, output_filename, header=header)



"""Functions / decorators for finding / requiring nibabel-data directory"""

import unittest
from os import environ, listdir
from os.path import dirname, exists, isdir, realpath
from os.path import join as pjoin


def get_nibabel_data():
    """Return path to nibabel-data or empty string if missing

    First use ``NIBABEL_DATA_DIR`` environment variable.

    If this variable is missing then look for data in directory below package
    directory.
    """
    nibabel_data = environ.get('NIBABEL_DATA_DIR')
    if nibabel_data is None:
        mod = __import__('nibabel')
        containing_path = dirname(dirname(realpath(mod.__file__)))
        nibabel_data = pjoin(containing_path, 'nibabel-data')
    return nibabel_data if isdir(nibabel_data) else ''


def needs_nibabel_data(subdir=None):
    """Decorator for tests needing nibabel-data

    Parameters
    ----------
    subdir : None or str
        Subdirectory we need in nibabel-data directory.  If None, only require
        nibabel-data directory itself.

    Returns
    -------
    skip_dec : decorator
        Decorator skipping tests if required directory not present
    """
    nibabel_data = get_nibabel_data()
    if nibabel_data == '':
        return unittest.skip('Need nibabel-data directory for this test')
    if subdir is None:
        return lambda x: x
    required_path = pjoin(nibabel_data, subdir)
    # Path should not be empty (as is the case for not-updated submodules)
    have_files = exists(required_path) and len(listdir(required_path)) > 0
    return unittest.skipUnless(have_files, f'Need files in {required_path} for these tests')



import copy
import os
import sys
import unittest
from io import BytesIO
from os.path import join as pjoin

import numpy as np
import pytest
from numpy.testing import assert_array_equal

from ...testing import assert_arr_dict_equal, clear_and_catch_warnings, data_path, error_warnings
from .. import trk as trk_module
from ..header import Field
from ..tractogram import Tractogram
from ..tractogram_file import HeaderError, HeaderWarning
from ..trk import (
    TrkFile,
    decode_value_from_name,
    encode_value_in_name,
    get_affine_trackvis_to_rasmm,
)
from .test_tractogram import assert_tractogram_equal

DATA = {}


def setup_module():
    global DATA

    DATA['empty_trk_fname'] = pjoin(data_path, 'empty.trk')
    # simple.trk contains only streamlines
    DATA['simple_trk_fname'] = pjoin(data_path, 'simple.trk')
    # standard.trk contains only streamlines
    DATA['standard_trk_fname'] = pjoin(data_path, 'standard.trk')
    # standard.LPS.trk contains only streamlines
    DATA['standard_LPS_trk_fname'] = pjoin(data_path, 'standard.LPS.trk')

    # complex.trk contains streamlines, scalars and properties
    DATA['complex_trk_fname'] = pjoin(data_path, 'complex.trk')
    DATA['complex_trk_big_endian_fname'] = pjoin(data_path, 'complex_big_endian.trk')

    DATA['streamlines'] = [
        np.arange(1 * 3, dtype='f4').reshape((1, 3)),
        np.arange(2 * 3, dtype='f4').reshape((2, 3)),
        np.arange(5 * 3, dtype='f4').reshape((5, 3)),
    ]

    DATA['fa'] = [
        np.array([[0.2]], dtype='f4'),
        np.array([[0.3], [0.4]], dtype='f4'),
        np.array([[0.5], [0.6], [0.6], [0.7], [0.8]], dtype='f4'),
    ]

    DATA['colors'] = [
        np.array([(1, 0, 0)] * 1, dtype='f4'),
        np.array([(0, 1, 0)] * 2, dtype='f4'),
        np.array([(0, 0, 1)] * 5, dtype='f4'),
    ]

    DATA['mean_curvature'] = [
        np.array([1.11], dtype='f4'),
        np.array([2.11], dtype='f4'),
        np.array([3.11], dtype='f4'),
    ]

    DATA['mean_torsion'] = [
        np.array([1.22], dtype='f4'),
        np.array([2.22], dtype='f4'),
        np.array([3.22], dtype='f4'),
    ]

    DATA['mean_colors'] = [
        np.array([1, 0, 0], dtype='f4'),
        np.array([0, 1, 0], dtype='f4'),
        np.array([0, 0, 1], dtype='f4'),
    ]

    DATA['data_per_point'] = {'colors': DATA['colors'], 'fa': DATA['fa']}
    DATA['data_per_streamline'] = {
        'mean_curvature': DATA['mean_curvature'],
        'mean_torsion': DATA['mean_torsion'],
        'mean_colors': DATA['mean_colors'],
    }

    DATA['empty_tractogram'] = Tractogram(affine_to_rasmm=np.eye(4))
    DATA['simple_tractogram'] = Tractogram(DATA['streamlines'], affine_to_rasmm=np.eye(4))
    DATA['complex_tractogram'] = Tractogram(
        DATA['streamlines'],
        DATA['data_per_streamline'],
        DATA['data_per_point'],
        affine_to_rasmm=np.eye(4),
    )


class TestTRK(unittest.TestCase):
    def test_load_empty_file(self):
        for lazy_load in [False, True]:
            trk = TrkFile.load(DATA['empty_trk_fname'], lazy_load=lazy_load)
            with pytest.warns(Warning) if lazy_load else error_warnings():
                assert_tractogram_equal(trk.tractogram, DATA['empty_tractogram'])

    def test_load_simple_file(self):
        for lazy_load in [False, True]:
            trk = TrkFile.load(DATA['simple_trk_fname'], lazy_load=lazy_load)
            with pytest.warns(Warning) if lazy_load else error_warnings():
                assert_tractogram_equal(trk.tractogram, DATA['simple_tractogram'])

    def test_load_complex_file(self):
        for lazy_load in [False, True]:
            trk = TrkFile.load(DATA['complex_trk_fname'], lazy_load=lazy_load)
            with pytest.warns(Warning) if lazy_load else error_warnings():
                assert_tractogram_equal(trk.tractogram, DATA['complex_tractogram'])

    def trk_with_bytes(self, trk_key='simple_trk_fname', endian='<'):
        """Return example trk file bytes and struct view onto bytes"""
        with open(DATA[trk_key], 'rb') as fobj:
            trk_bytes = bytearray(fobj.read())
        dt = trk_module.header_2_dtype.newbyteorder(endian)
        trk_struct = np.ndarray((1,), dt, buffer=trk_bytes)
        trk_struct.flags.writeable = True
        return trk_struct, trk_bytes

    def test_load_file_with_wrong_information(self):
        # Simulate a TRK file where `voxel_order` is lowercase.
        trk_struct1, trk_bytes1 = self.trk_with_bytes()
        trk_struct1[Field.VOXEL_ORDER] = b'LAS'
        trk1 = TrkFile.load(BytesIO(trk_bytes1))
        trk_struct2, trk_bytes2 = self.trk_with_bytes()
        trk_struct2[Field.VOXEL_ORDER] = b'las'
        trk2 = TrkFile.load(BytesIO(trk_bytes2))
        trk1_aff2rasmm = get_affine_trackvis_to_rasmm(trk1.header)
        trk2_aff2rasmm = get_affine_trackvis_to_rasmm(trk2.header)
        assert_array_equal(trk1_aff2rasmm, trk2_aff2rasmm)

        # Simulate a TRK file where `count` was not provided.
        trk_struct, trk_bytes = self.trk_with_bytes()
        trk_struct[Field.NB_STREAMLINES] = 0
        trk = TrkFile.load(BytesIO(trk_bytes), lazy_load=False)
        assert_tractogram_equal(trk.tractogram, DATA['simple_tractogram'])

        # Simulate a TRK where `vox_to_ras` is not recorded (i.e. all zeros).
        trk_struct, trk_bytes = self.trk_with_bytes()
        trk_struct[Field.VOXEL_TO_RASMM] = np.zeros((4, 4))
        with pytest.warns(HeaderWarning, match='identity'):
            trk = TrkFile.load(BytesIO(trk_bytes))
        assert_array_equal(trk.affine, np.eye(4))

        # Simulate a TRK where `vox_to_ras` is invalid.
        trk_struct, trk_bytes = self.trk_with_bytes()
        trk_struct[Field.VOXEL_TO_RASMM] = np.diag([0, 0, 0, 1])
        with clear_and_catch_warnings(modules=[trk_module]):
            with pytest.raises(HeaderError):
                TrkFile.load(BytesIO(trk_bytes))

        # Simulate a TRK file where `voxel_order` was not provided.
        trk_struct, trk_bytes = self.trk_with_bytes()
        trk_struct[Field.VOXEL_ORDER] = b''
        with pytest.warns(HeaderWarning, match='LPS'):
            TrkFile.load(BytesIO(trk_bytes))

        # Simulate a TRK file with an unsupported version.
        trk_struct, trk_bytes = self.trk_with_bytes()
        trk_struct['version'] = 123
        with pytest.raises(HeaderError):
            TrkFile.load(BytesIO(trk_bytes))

        # Simulate a TRK file with a wrong hdr_size.
        trk_struct, trk_bytes = self.trk_with_bytes()
        trk_struct['hdr_size'] = 1234
        with pytest.raises(HeaderError):
            TrkFile.load(BytesIO(trk_bytes))

        # Simulate a TRK file with a wrong scalar_name.
        trk_struct, trk_bytes = self.trk_with_bytes('complex_trk_fname')
        trk_struct['scalar_name'][0, 0] = b'colors\x003\x004'
        with pytest.raises(HeaderError):
            TrkFile.load(BytesIO(trk_bytes))

        # Simulate a TRK file with a wrong property_name.
        trk_struct, trk_bytes = self.trk_with_bytes('complex_trk_fname')
        trk_struct['property_name'][0, 0] = b'colors\x003\x004'
        with pytest.raises(HeaderError):
            TrkFile.load(BytesIO(trk_bytes))

    def test_load_trk_version_1(self):
        # Simulate and test a TRK (version 1).
        # First check that setting the RAS affine works in version 2.
        trk_struct, trk_bytes = self.trk_with_bytes()
        trk_struct[Field.VOXEL_TO_RASMM] = np.diag([2, 3, 4, 1])
        trk = TrkFile.load(BytesIO(trk_bytes))
        assert_array_equal(trk.affine, np.diag([2, 3, 4, 1]))
        # Next check that affine assumed identity if version 1.
        trk_struct['version'] = 1
        with pytest.warns(HeaderWarning, match='identity'):
            trk = TrkFile.load(BytesIO(trk_bytes))
        assert_array_equal(trk.affine, np.eye(4))
        assert_array_equal(trk.header['version'], 1)

    def test_load_complex_file_in_big_endian(self):
        trk_struct, trk_bytes = self.trk_with_bytes('complex_trk_big_endian_fname', endian='>')
        # We use hdr_size as an indicator of little vs big endian.
        good_orders = '>' if sys.byteorder == 'little' else '>='
        hdr_size = trk_struct['hdr_size']
        assert hdr_size.dtype.byteorder in good_orders
        assert hdr_size == 1000

        for lazy_load in [False, True]:
            trk = TrkFile.load(DATA['complex_trk_big_endian_fname'], lazy_load=lazy_load)
            with pytest.warns(Warning) if lazy_load else error_warnings():
                assert_tractogram_equal(trk.tractogram, DATA['complex_tractogram'])

    def test_tractogram_file_properties(self):
        trk = TrkFile.load(DATA['simple_trk_fname'])
        assert trk.streamlines == trk.tractogram.streamlines
        assert_array_equal(trk.affine, trk.header[Field.VOXEL_TO_RASMM])

    def test_write_empty_file(self):
        tractogram = Tractogram(affine_to_rasmm=np.eye(4))

        trk_file = BytesIO()
        trk = TrkFile(tractogram)
        trk.save(trk_file)
        trk_file.seek(0, os.SEEK_SET)

        new_trk = TrkFile.load(trk_file)
        assert_tractogram_equal(new_trk.tractogram, tractogram)

        new_trk_orig = TrkFile.load(DATA['empty_trk_fname'])
        assert_tractogram_equal(new_trk.tractogram, new_trk_orig.tractogram)

        trk_file.seek(0, os.SEEK_SET)
        assert trk_file.read() == open(DATA['empty_trk_fname'], 'rb').read()

    def test_write_simple_file(self):
        tractogram = Tractogram(DATA['streamlines'], affine_to_rasmm=np.eye(4))

        trk_file = BytesIO()
        trk = TrkFile(tractogram)
        trk.save(trk_file)
        trk_file.seek(0, os.SEEK_SET)

        new_trk = TrkFile.load(trk_file)
        assert_tractogram_equal(new_trk.tractogram, tractogram)

        new_trk_orig = TrkFile.load(DATA['simple_trk_fname'])
        assert_tractogram_equal(new_trk.tractogram, new_trk_orig.tractogram)

        trk_file.seek(0, os.SEEK_SET)
        assert trk_file.read() == open(DATA['simple_trk_fname'], 'rb').read()

    def test_write_complex_file(self):
        # With scalars
        tractogram = Tractogram(
            DATA['streamlines'], data_per_point=DATA['data_per_point'], affine_to_rasmm=np.eye(4)
        )

        trk_file = BytesIO()
        trk = TrkFile(tractogram)
        trk.save(trk_file)
        trk_file.seek(0, os.SEEK_SET)

        new_trk = TrkFile.load(trk_file, lazy_load=False)
        assert_tractogram_equal(new_trk.tractogram, tractogram)

        # With properties
        data_per_streamline = DATA['data_per_streamline']
        tractogram = Tractogram(
            DATA['streamlines'], data_per_streamline=data_per_streamline, affine_to_rasmm=np.eye(4)
        )

        trk = TrkFile(tractogram)
        trk_file = BytesIO()
        trk.save(trk_file)
        trk_file.seek(0, os.SEEK_SET)

        new_trk = TrkFile.load(trk_file, lazy_load=False)
        assert_tractogram_equal(new_trk.tractogram, tractogram)

        # With scalars and properties
        data_per_streamline = DATA['data_per_streamline']
        tractogram = Tractogram(
            DATA['streamlines'],
            data_per_point=DATA['data_per_point'],
            data_per_streamline=data_per_streamline,
            affine_to_rasmm=np.eye(4),
        )

        trk_file = BytesIO()
        trk = TrkFile(tractogram)
        trk.save(trk_file)
        trk_file.seek(0, os.SEEK_SET)

        new_trk = TrkFile.load(trk_file, lazy_load=False)
        assert_tractogram_equal(new_trk.tractogram, tractogram)

        new_trk_orig = TrkFile.load(DATA['complex_trk_fname'])
        assert_tractogram_equal(new_trk.tractogram, new_trk_orig.tractogram)

        trk_file.seek(0, os.SEEK_SET)
        assert trk_file.read() == open(DATA['complex_trk_fname'], 'rb').read()

    def test_load_write_file(self):
        for fname in [
            DATA['empty_trk_fname'],
            DATA['simple_trk_fname'],
            DATA['complex_trk_fname'],
        ]:
            for lazy_load in [False, True]:
                trk = TrkFile.load(fname, lazy_load=lazy_load)
                trk_file = BytesIO()
                trk.save(trk_file)

                new_trk = TrkFile.load(fname, lazy_load=False)
                assert_tractogram_equal(new_trk.tractogram, trk.tractogram)

    def test_load_write_LPS_file(self):
        # Load the RAS and LPS version of the standard.
        trk_RAS = TrkFile.load(DATA['standard_trk_fname'], lazy_load=False)
        trk_LPS = TrkFile.load(DATA['standard_LPS_trk_fname'], lazy_load=False)
        assert_tractogram_equal(trk_LPS.tractogram, trk_RAS.tractogram)

        # Write back the standard.
        trk_file = BytesIO()
        trk = TrkFile(trk_LPS.tractogram, trk_LPS.header)
        trk.save(trk_file)
        trk_file.seek(0, os.SEEK_SET)

        new_trk = TrkFile.load(trk_file)

        assert_arr_dict_equal(new_trk.header, trk.header)
        assert_tractogram_equal(new_trk.tractogram, trk.tractogram)

        new_trk_orig = TrkFile.load(DATA['standard_LPS_trk_fname'])
        assert_tractogram_equal(new_trk.tractogram, new_trk_orig.tractogram)

        trk_file.seek(0, os.SEEK_SET)
        assert trk_file.read() == open(DATA['standard_LPS_trk_fname'], 'rb').read()

        # Test writing a file where the header is missing the
        # Field.VOXEL_ORDER.
        trk_file = BytesIO()

        # For TRK file format, the default voxel order is LPS.
        header = copy.deepcopy(trk_LPS.header)
        header[Field.VOXEL_ORDER] = b''

        trk = TrkFile(trk_LPS.tractogram, header)
        trk.save(trk_file)
        trk_file.seek(0, os.SEEK_SET)

        new_trk = TrkFile.load(trk_file)

        assert_arr_dict_equal(new_trk.header, trk_LPS.header)
        assert_tractogram_equal(new_trk.tractogram, trk.tractogram)

        new_trk_orig = TrkFile.load(DATA['standard_LPS_trk_fname'])
        assert_tractogram_equal(new_trk.tractogram, new_trk_orig.tractogram)

        trk_file.seek(0, os.SEEK_SET)
        assert trk_file.read() == open(DATA['standard_LPS_trk_fname'], 'rb').read()

    def test_write_optional_header_fields(self):
        # The TRK file format doesn't support additional header fields.
        # If provided, they will be ignored.
        tractogram = Tractogram(affine_to_rasmm=np.eye(4))

        trk_file = BytesIO()
        header = {'extra': 1234}
        trk = TrkFile(tractogram, header)
        trk.save(trk_file)
        trk_file.seek(0, os.SEEK_SET)

        new_trk = TrkFile.load(trk_file)
        assert 'extra' not in new_trk.header

    def test_write_too_many_scalars_and_properties(self):
        # TRK supports up to 10 data_per_point.
        data_per_point = {}
        for i in range(10):
            data_per_point[f'#{i}'] = DATA['fa']

            tractogram = Tractogram(
                DATA['streamlines'], data_per_point=data_per_point, affine_to_rasmm=np.eye(4)
            )

            trk_file = BytesIO()
            trk = TrkFile(tractogram)
            trk.save(trk_file)
            trk_file.seek(0, os.SEEK_SET)

            new_trk = TrkFile.load(trk_file, lazy_load=False)
            assert_tractogram_equal(new_trk.tractogram, tractogram)

        # More than 10 data_per_point should raise an error.
        data_per_point[f'#{i + 1}'] = DATA['fa']

        tractogram = Tractogram(
            DATA['streamlines'], data_per_point=data_per_point, affine_to_rasmm=np.eye(4)
        )

        trk = TrkFile(tractogram)
        with pytest.raises(ValueError):
            trk.save(BytesIO())

        # TRK supports up to 10 data_per_streamline.
        data_per_streamline = {}
        for i in range(10):
            data_per_streamline[f'#{i}'] = DATA['mean_torsion']

            tractogram = Tractogram(
                DATA['streamlines'],
                data_per_streamline=data_per_streamline,
                affine_to_rasmm=np.eye(4),
            )

            trk_file = BytesIO()
            trk = TrkFile(tractogram)
            trk.save(trk_file)
            trk_file.seek(0, os.SEEK_SET)

            new_trk = TrkFile.load(trk_file, lazy_load=False)
            assert_tractogram_equal(new_trk.tractogram, tractogram)

        # More than 10 data_per_streamline should raise an error.
        data_per_streamline[f'#{i + 1}'] = DATA['mean_torsion']

        tractogram = Tractogram(DATA['streamlines'], data_per_streamline=data_per_streamline)

        trk = TrkFile(tractogram)
        with pytest.raises(ValueError):
            trk.save(BytesIO())

    def test_write_scalars_and_properties_name_too_long(self):
        # TRK supports data_per_point name up to 20 characters.
        # However, we reserve the last two characters to store
        # the number of values associated to each data_per_point.
        # So in reality we allow name of 18 characters, otherwise
        # the name is truncated and warning is issue.
        for nb_chars in range(22):
            data_per_point = {'A' * nb_chars: DATA['colors']}
            tractogram = Tractogram(
                DATA['streamlines'], data_per_point=data_per_point, affine_to_rasmm=np.eye(4)
            )

            trk = TrkFile(tractogram)
            if nb_chars > 18:
                with pytest.raises(ValueError):
                    trk.save(BytesIO())
            else:
                trk.save(BytesIO())

            data_per_point = {'A' * nb_chars: DATA['fa']}
            tractogram = Tractogram(
                DATA['streamlines'], data_per_point=data_per_point, affine_to_rasmm=np.eye(4)
            )

            trk = TrkFile(tractogram)
            if nb_chars > 20:
                with pytest.raises(ValueError):
                    trk.save(BytesIO())
            else:
                trk.save(BytesIO())

        # TRK supports data_per_streamline name up to 20 characters.
        # However, we reserve the last two characters to store
        # the number of values associated to each data_per_streamline.
        # So in reality we allow name of 18 characters, otherwise
        # the name is truncated and warning is issue.
        for nb_chars in range(22):
            data_per_streamline = {'A' * nb_chars: DATA['mean_colors']}
            tractogram = Tractogram(
                DATA['streamlines'],
                data_per_streamline=data_per_streamline,
                affine_to_rasmm=np.eye(4),
            )

            trk = TrkFile(tractogram)
            if nb_chars > 18:
                with pytest.raises(ValueError):
                    trk.save(BytesIO())
            else:
                trk.save(BytesIO())

            data_per_streamline = {'A' * nb_chars: DATA['mean_torsion']}
            tractogram = Tractogram(
                DATA['streamlines'],
                data_per_streamline=data_per_streamline,
                affine_to_rasmm=np.eye(4),
            )

            trk = TrkFile(tractogram)
            if nb_chars > 20:
                with pytest.raises(ValueError):
                    trk.save(BytesIO())
            else:
                trk.save(BytesIO())

    def test_str(self):
        trk = TrkFile.load(DATA['complex_trk_fname'])
        str(trk)  # Simply test it's not failing when called.

    def test_header_read_restore(self):
        # Test that reading a header restores the file position
        trk_fname = DATA['simple_trk_fname']
        bio = BytesIO()
        bio.write(b'Along my very merry way')
        hdr_pos = bio.tell()
        hdr_from_fname = TrkFile._read_header(trk_fname)
        with open(trk_fname, 'rb') as fobj:
            bio.write(fobj.read())
        bio.seek(hdr_pos)
        # Check header is as expected
        hdr_from_fname['_offset_data'] += hdr_pos  # Correct for start position
        assert_arr_dict_equal(TrkFile._read_header(bio), hdr_from_fname)
        # Check fileobject file position has not changed
        assert bio.tell() == hdr_pos


def test_encode_names():
    # Test function for encoding numbers into property names
    b0 = b'\x00'
    assert encode_value_in_name(0, 'foo', 10) == b'foo' + b0 * 7
    assert encode_value_in_name(1, 'foo', 10) == b'foo' + b0 * 7
    assert encode_value_in_name(8, 'foo', 10) == b'foo' + b0 + b'8' + b0 * 5
    assert encode_value_in_name(40, 'foobar', 10) == b'foobar' + b0 + b'40' + b0
    assert encode_value_in_name(1, 'foobarbazz', 10) == b'foobarbazz'

    with pytest.raises(ValueError):
        encode_value_in_name(1, 'foobarbazzz', 10)

    with pytest.raises(ValueError):
        encode_value_in_name(2, 'foobarbazzz', 10)

    assert encode_value_in_name(2, 'foobarba', 10) == b'foobarba\x002'


def test_decode_names():
    # Test function for decoding name string into name, number
    b0 = b'\x00'
    assert decode_value_from_name(b'') == ('', 0)
    assert decode_value_from_name(b'foo' + b0 * 7) == ('foo', 1)
    assert decode_value_from_name(b'foo\x008' + b0 * 5) == ('foo', 8)
    assert decode_value_from_name(b'foobar\x0010\x00') == ('foobar', 10)

    with pytest.raises(ValueError):
        decode_value_from_name(b'foobar\x0010\x01')

    with pytest.raises(HeaderError):
        decode_value_from_name(b'foo\x0010\x00111')



"""Data packages for DICOM testing"""

from ... import data as nibd

PUBLIC_PKG_DEF = dict(relpath='nipy/dicom/public', name='nipy-dicom-public', version='0.1')

PRIVATE_PKG_DEF = dict(relpath='nipy/dicom/private', name='nipy-dicom-private', version='0.1')


PUBLIC_DS = nibd.datasource_or_bomber(PUBLIC_PKG_DEF)
PRIVATE_DS = nibd.datasource_or_bomber(PRIVATE_PKG_DEF)



# emacs: -*- mode: python-mode; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the NiBabel package for the
#   copyright and license terms.
#
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Testing for orientations module"""

import numpy as np
import pytest
from numpy.testing import assert_array_equal

from ..affines import from_matvec, to_matvec
from ..nifti1 import Nifti1Image
from ..orientations import (
    OrientationError,
    aff2axcodes,
    apply_orientation,
    axcodes2ornt,
    flip_axis,
    inv_ornt_aff,
    io_orientation,
    ornt2axcodes,
    ornt_transform,
)
from ..testing import deprecated_to, expires

IN_ARRS = [
    np.eye(4),
    [
        [0, 0, 1, 0],
        [0, 1, 0, 0],
        [1, 0, 0, 0],
        [0, 0, 0, 1],
    ],
    [
        [0, 1, 0, 0],
        [0, 0, 1, 0],
        [1, 0, 0, 0],
        [0, 0, 0, 1],
    ],
    [
        [3, 1, 0, 0],
        [1, 3, 0, 0],
        [0, 0, 1, 0],
        [0, 0, 0, 1],
    ],
    [
        [1, 3, 0, 0],
        [3, 1, 0, 0],
        [0, 0, 1, 0],
        [0, 0, 0, 1],
    ],
]

OUT_ORNTS = [
    [
        [0, 1],
        [1, 1],
        [2, 1],
    ],
    [
        [2, 1],
        [1, 1],
        [0, 1],
    ],
    [
        [2, 1],
        [0, 1],
        [1, 1],
    ],
    [
        [0, 1],
        [1, 1],
        [2, 1],
    ],
    [
        [1, 1],
        [0, 1],
        [2, 1],
    ],
]

IN_ARRS.extend(
    [
        [np.cos(np.pi / 6 + i * np.pi / 2), np.sin(np.pi / 6 + i * np.pi / 2), 0, 0],
        [-np.sin(np.pi / 6 + i * np.pi / 2), np.cos(np.pi / 6 + i * np.pi / 2), 0, 0],
        [0, 0, 1, 0],
        [0, 0, 0, 1],
    ]
    for i in range(4)
)

OUT_ORNTS += [
    [
        [0, 1],
        [1, 1],
        [2, 1],
    ],
    [
        [1, -1],
        [0, 1],
        [2, 1],
    ],
    [
        [0, -1],
        [1, -1],
        [2, 1],
    ],
    [
        [1, 1],
        [0, -1],
        [2, 1],
    ],
]


IN_ARRS = [np.array(arr) for arr in IN_ARRS]
OUT_ORNTS = [np.array(ornt) for ornt in OUT_ORNTS]


_LABELS = ['RL', 'AP', 'SI']
ALL_AXCODES = [
    (_LABELS[i0][j0], _LABELS[i1][j1], _LABELS[i2][j2])
    for i0 in range(3)
    for i1 in range(3)
    for i2 in range(3)
    if i0 != i1 != i2 != i0
    for j0 in range(2)
    for j1 in range(2)
    for j2 in range(2)
]

ALL_ORNTS = [
    [[i0, j0], [i1, j1], [i2, j2]]
    for i0 in range(3)
    for i1 in range(3)
    for i2 in range(3)
    if i0 != i1 != i2 != i0
    for j0 in [1, -1]
    for j1 in [1, -1]
    for j2 in [1, -1]
]


def same_transform(taff, ornt, shape):
    # Applying transformations implied by `ornt` to a made-up array
    # ``arr`` of shape `shape`, results in ``t_arr``. When the point
    # indices from ``arr`` are transformed by (the inverse of) `taff`,
    # and we index into ``t_arr`` with these transformed points, then we
    # should get the same values as we would from indexing into arr with
    # the untransformed points.
    shape = np.array(shape)
    size = np.prod(shape)
    arr = np.arange(size).reshape(shape)
    # apply ornt transformations
    t_arr = apply_orientation(arr, ornt)
    # get all point indices in arr
    i, j, k = shape
    arr_pts = np.mgrid[:i, :j, :k].reshape((3, -1))
    # inverse of taff takes us from point index in arr to point index in
    # t_arr
    itaff = np.linalg.inv(taff)
    # apply itaff so that points indexed in t_arr should correspond
    o2t_pts = np.dot(itaff[:3, :3], arr_pts) + itaff[:3, 3][:, None]
    assert np.allclose(np.round(o2t_pts), o2t_pts)
    # fancy index out the t_arr values
    vals = t_arr[tuple(o2t_pts.astype('i'))]
    return np.all(vals == arr.ravel())


def test_apply():
    # most tests are in ``same_transform`` above, via the
    # test_io_orientations
    a = np.arange(24).reshape((2, 3, 4))
    # Test 4D with an example orientation
    ornt = OUT_ORNTS[-1]
    t_arr = apply_orientation(a[:, :, :, None], ornt)
    assert t_arr.ndim == 4
    # Orientation errors
    with pytest.raises(OrientationError):
        apply_orientation(a[:, :, 1], ornt)
    with pytest.raises(OrientationError):
        apply_orientation(a, [[0, 1], [np.nan, np.nan], [2, 1]])
    for ornt in ALL_ORNTS:
        t_arr = apply_orientation(a, ornt)
        assert_array_equal(a.shape, np.array(t_arr.shape)[np.array(ornt)[:, 0]])


def test_io_orientation():
    for shape in ((2, 3, 4), (20, 15, 7)):
        for in_arr, out_ornt in zip(IN_ARRS, OUT_ORNTS):
            ornt = io_orientation(in_arr)
            assert_array_equal(ornt, out_ornt)
            taff = inv_ornt_aff(ornt, shape)
            assert same_transform(taff, ornt, shape)
            for axno in range(3):
                arr = in_arr.copy()
                ex_ornt = out_ornt.copy()
                # flip the input axis in affine
                arr[:, axno] *= -1
                # check that result shows flip
                ex_ornt[axno, 1] *= -1
                ornt = io_orientation(arr)
                assert_array_equal(ornt, ex_ornt)
                taff = inv_ornt_aff(ornt, shape)
                assert same_transform(taff, ornt, shape)
    # Test nasty hang for zero columns
    rzs = np.c_[np.diag([2, 3, 4, 5]), np.zeros((4, 3))]
    arr = from_matvec(rzs, [15, 16, 17, 18])
    ornt = io_orientation(arr)
    assert_array_equal(
        ornt,
        [
            [0, 1],
            [1, 1],
            [2, 1],
            [3, 1],
            [np.nan, np.nan],
            [np.nan, np.nan],
            [np.nan, np.nan],
        ],
    )
    # Test behavior of thresholding
    def_aff = np.array(
        [
            [1.0, 1, 0, 0],
            [0, 0, 0, 0],
            [0, 0, 1, 0],
            [0, 0, 0, 1],
        ]
    )
    fail_tol = np.array(
        [
            [0, 1],
            [np.nan, np.nan],
            [2, 1],
        ]
    )
    pass_tol = np.array(
        [
            [0, 1],
            [1, 1],
            [2, 1],
        ]
    )
    eps = np.finfo(float).eps
    # Test that a Y axis appears as we increase the difference between the
    # first two columns
    for y_val, has_y in (
        (0, False),
        (eps, False),
        (eps * 5, False),
        (eps * 10, True),
    ):
        def_aff[1, 1] = y_val
        res = pass_tol if has_y else fail_tol
        assert_array_equal(io_orientation(def_aff), res)
    # Test tol input argument
    def_aff[1, 1] = eps
    assert_array_equal(io_orientation(def_aff, tol=0), pass_tol)
    def_aff[1, 1] = eps * 10
    assert_array_equal(io_orientation(def_aff, tol=1e-5), fail_tol)
    # Test drop of rows, columns
    mat, vec = to_matvec(def_aff)
    aff_extra_col = np.zeros((4, 5))
    aff_extra_col[-1, -1] = 1  # Not strictly necessary, but for completeness
    aff_extra_col[:3, :3] = mat
    aff_extra_col[:3, -1] = vec
    assert_array_equal(
        io_orientation(aff_extra_col, tol=1e-5),
        [
            [np.nan, np.nan],
            [0, 1],
            [2, 1],
            [np.nan, np.nan],
        ],
    )
    aff_extra_row = np.zeros((5, 4))
    aff_extra_row[-1, -1] = 1  # Not strictly necessary, but for completeness
    aff_extra_row[:3, :3] = mat
    aff_extra_row[:3, -1] = vec
    assert_array_equal(
        io_orientation(aff_extra_row, tol=1e-5),
        [
            [0, 1],
            [np.nan, np.nan],
            [2, 1],
        ],
    )


def test_io_orientation_column_strength_regression():
    # Real-world IAR affine that triggered this test.
    # The feature of interest is that the third row of both the first and third columns
    # has the highest magnitude.
    # An left-to-right labeling will change labels when the columns are permuted,
    # hence we use a magnitude-based labeling for consistency.
    # See https://github.com/nipy/nibabel/issues/1449 for further discussion
    affine = np.array(
        [
            [2.08759499e00, 7.70245194e-02, 1.12271041e-01, -1.67531357e01],
            [-1.04219818e00, 1.58019245e-01, -5.34135476e-02, 1.22405062e01],
            [-2.33361936e00, -1.66752085e-03, 1.24289364e-01, -1.09963446e01],
            [0.00000000e00, 0.00000000e00, 0.00000000e00, 1.00000000e00],
        ]
    )

    # Verify IAR orientation
    iar_ornt = io_orientation(affine)
    assert_array_equal(iar_ornt, axcodes2ornt('IAR'))

    # Reorient to RAS and verify it is identified as RAS
    img = Nifti1Image(np.zeros((2, 3, 4), dtype=np.float32), affine)
    img_ras = img.as_reoriented(iar_ornt)
    img_ras_ornt = io_orientation(img_ras.affine)

    # Using the previous method, this appeared to be SAL
    assert_array_equal(img_ras_ornt, axcodes2ornt('RAS'))
    assert img_ras.shape == (4, 3, 2)

    # Test idempotence
    img_ras_reornt = img_ras.as_reoriented(img_ras_ornt)
    img_ras_reornt_ornt = io_orientation(img_ras_reornt.affine)

    assert_array_equal(img_ras_reornt_ornt, axcodes2ornt('RAS'))
    assert img_ras_reornt.shape == (4, 3, 2)


def test_ornt_transform():
    assert_array_equal(
        ornt_transform(
            [[0, 1], [1, 1], [2, -1]],
            [[1, 1], [0, 1], [2, 1]],
        ),
        [[1, 1], [0, 1], [2, -1]],
    )
    assert_array_equal(
        ornt_transform(
            [[0, 1], [1, 1], [2, 1]],
            [[2, 1], [0, -1], [1, 1]],
        ),
        [[1, -1], [2, 1], [0, 1]],
    )
    # Must have same shape
    with pytest.raises(ValueError):
        ornt_transform([[0, 1], [1, 1]], [[0, 1], [1, 1], [2, 1]])

    # Must be (N,2) in shape
    with pytest.raises(ValueError):
        ornt_transform(
            [[0, 1, 1], [1, 1, 1]],
            [[0, 1, 1], [1, 1, 1]],
        )

    # Target axes must exist in source
    with pytest.raises(ValueError):
        ornt_transform(
            [[0, 1], [1, 1], [1, 1]],
            [[0, 1], [1, 1], [2, 1]],
        )


def test_ornt2axcodes():
    # Recoding orientation to axis codes
    labels = (('left', 'right'), ('back', 'front'), ('down', 'up'))
    assert ornt2axcodes([[0, 1], [1, 1], [2, 1]], labels) == ('right', 'front', 'up')
    assert ornt2axcodes([[0, -1], [1, -1], [2, -1]], labels) == ('left', 'back', 'down')
    assert ornt2axcodes([[2, -1], [1, -1], [0, -1]], labels) == ('down', 'back', 'left')
    assert ornt2axcodes([[1, 1], [2, -1], [0, 1]], labels) == ('front', 'down', 'right')
    # default is RAS output directions
    assert ornt2axcodes([[0, 1], [1, 1], [2, 1]]) == ('R', 'A', 'S')
    # dropped axes produce None
    assert ornt2axcodes([[0, 1], [np.nan, np.nan], [2, 1]]) == ('R', None, 'S')
    # Non integer axes raises error
    with pytest.raises(ValueError):
        ornt2axcodes([[0.1, 1]])
    # As do directions not in range
    with pytest.raises(ValueError):
        ornt2axcodes([[0, 0]])

    for axcodes, ornt in zip(ALL_AXCODES, ALL_ORNTS):
        assert ornt2axcodes(ornt) == axcodes


def test_axcodes2ornt():
    # Go from axcodes back to orientations
    labels = (('left', 'right'), ('back', 'front'), ('down', 'up'))
    assert_array_equal(axcodes2ornt(('right', 'front', 'up'), labels), [[0, 1], [1, 1], [2, 1]])
    assert_array_equal(axcodes2ornt(('left', 'back', 'down'), labels), [[0, -1], [1, -1], [2, -1]])
    assert_array_equal(axcodes2ornt(('down', 'back', 'left'), labels), [[2, -1], [1, -1], [0, -1]])
    assert_array_equal(axcodes2ornt(('front', 'down', 'right'), labels), [[1, 1], [2, -1], [0, 1]])

    # default is RAS output directions
    default = np.c_[range(3), [1] * 3]
    assert_array_equal(axcodes2ornt(('R', 'A', 'S')), default)

    # dropped axes produce None
    assert_array_equal(axcodes2ornt(('R', None, 'S')), [[0, 1], [np.nan, np.nan], [2, 1]])

    # Missing axcodes raise an error
    assert_array_equal(axcodes2ornt('RAS'), default)
    with pytest.raises(ValueError):
        axcodes2ornt('rAS')
    # None is OK as axis code
    assert_array_equal(axcodes2ornt(('R', None, 'S')), [[0, 1], [np.nan, np.nan], [2, 1]])
    # Bad axis code with None also raises error.
    with pytest.raises(ValueError):
        axcodes2ornt(('R', None, 's'))
    # Axis codes checked with custom labels
    labels = ('SD', 'BF', 'lh')
    assert_array_equal(axcodes2ornt('BlD', labels), [[1, -1], [2, -1], [0, 1]])
    with pytest.raises(ValueError):
        axcodes2ornt('blD', labels)

    # Duplicate labels
    for labels in [('SD', 'BF', 'lD'), ('SD', 'SF', 'lD')]:
        with pytest.raises(ValueError):
            axcodes2ornt('blD', labels)

    for axcodes, ornt in zip(ALL_AXCODES, ALL_ORNTS):
        assert_array_equal(axcodes2ornt(axcodes), ornt)


def test_aff2axcodes():
    assert aff2axcodes(np.eye(4)) == tuple('RAS')
    aff = [[0, 1, 0, 10], [-1, 0, 0, 20], [0, 0, 1, 30], [0, 0, 0, 1]]
    assert aff2axcodes(aff, (('L', 'R'), ('B', 'F'), ('D', 'U'))) == ('B', 'R', 'U')
    assert aff2axcodes(aff, (('L', 'R'), ('B', 'F'), ('D', 'U'))) == ('B', 'R', 'U')


def test_inv_ornt_aff():
    # Extra tests for inv_ornt_aff routines (also tested in
    # io_orientations test)
    with pytest.raises(OrientationError):
        inv_ornt_aff([[0, 1], [1, -1], [np.nan, np.nan]], (3, 4, 5))


@expires('5.0.0')
def test_flip_axis_deprecation():
    a = np.arange(24).reshape((2, 3, 4))
    axis = 1
    with deprecated_to('5.0.0'):
        a_flipped = flip_axis(a, axis)
    assert_array_equal(a_flipped, np.flip(a, axis))
